# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from collections import defaultdict

import pyvex
import claripy

from angr.utils.bits import s2u, u2s
from angr.block import Block
from angr.analyses.analysis import Analysis
from angr.analyses import AnalysesHub
from angr.knowledge_plugins.functions import Function
from angr.codenode import BlockNode, HookNode
from angr.engines.light import SimEngineNostmtVEX, SimEngineLight, SpOffset, RegisterOffset
from angr.calling_conventions import SimRegArg, SimStackArg, default_cc
from angr.sim_type import SimTypeBottom, SimTypeFunction
from angr.utils.types import dereference_simtype_by_lib
from .utils import is_sane_register_variable

if TYPE_CHECKING:
    from angr.codenode import CodeNode


class FactCollectorState:
    """
    The abstract state for FactCollector.
    """

    __slots__ = (
        "bp_value",
        "callee_stored_regs",
        "reg_reads",
        "reg_reads_count",
        "reg_writes",
        "simple_stack",
        "sp_value",
        "stack_reads",
        "stack_writes",
        "tmps",
    )

    def __init__(self):
        self.tmps = {}
        self.simple_stack = {}

        self.callee_stored_regs: dict[int, int] = {}  # reg offset -> stack offset
        self.reg_reads = {}
        self.reg_reads_count = defaultdict(int)
        self.reg_writes: set[int] = set()
        self.stack_reads = {}
        self.stack_writes: set[int] = set()
        self.sp_value = 0
        self.bp_value = 0

    def register_read(self, offset: int, size_in_bytes: int):
        self.reg_reads_count[offset] += 1
        if offset in self.reg_writes:
            return
        if offset not in self.reg_reads:
            self.reg_reads[offset] = size_in_bytes
        else:
            self.reg_reads[offset] = max(self.reg_reads[offset], size_in_bytes)

    def register_read_undo(self, offset: int) -> None:
        if offset not in self.reg_reads or offset not in self.reg_reads_count:
            return
        self.reg_reads_count[offset] -= 1
        if self.reg_reads_count[offset] == 0:
            self.reg_reads.pop(offset)
            self.reg_reads_count.pop(offset)

    def register_written(self, offset: int, size_in_bytes: int):
        for o in range(size_in_bytes):
            self.reg_writes.add(offset + o)

    def stack_read(self, offset: int, size_in_bytes: int):
        if offset in self.stack_writes:
            return
        if offset not in self.stack_reads:
            self.stack_reads[offset] = size_in_bytes
        else:
            self.stack_reads[offset] = max(self.stack_reads[offset], size_in_bytes)

    def stack_written(self, offset: int, size_int_bytes: int):
        for o in range(size_int_bytes):
            self.stack_writes.add(offset + o)

    def copy(self, with_tmps: bool = False) -> FactCollectorState:
        new_state = FactCollectorState()
        new_state.reg_reads = self.reg_reads.copy()
        new_state.stack_reads = self.stack_reads.copy()
        new_state.stack_writes = self.stack_writes.copy()
        new_state.reg_writes = self.reg_writes.copy()
        new_state.callee_stored_regs = self.callee_stored_regs.copy()
        new_state.sp_value = self.sp_value
        new_state.bp_value = self.bp_value
        new_state.simple_stack = self.simple_stack.copy()
        new_state.reg_reads_count = self.reg_reads_count.copy()
        if with_tmps:
            new_state.tmps = self.tmps.copy()
        return new_state


binop_handler = SimEngineNostmtVEX[FactCollectorState, claripy.ast.BV, FactCollectorState].binop_handler


class SimEngineFactCollectorVEX(
    SimEngineNostmtVEX[FactCollectorState, SpOffset | RegisterOffset | int, None],
    SimEngineLight[FactCollectorState, SpOffset | RegisterOffset | int, Block, None],
):
    """
    THe engine for FactCollector.
    """

    def __init__(self, project, bp_as_gpr: bool):
        self.bp_as_gpr = bp_as_gpr
        super().__init__(project)

    def _process_block_end(self, stmt_result: list, whitelist: set[int] | None) -> None:
        if self.block.vex.jumpkind == "Ijk_Call" and self.arch.ret_offset is not None:
            self.state.register_written(self.arch.ret_offset, self.arch.bytes)

    def _top(self, bits: int):  # type: ignore
        return None

    def _is_top(self, expr: Any) -> bool:
        raise NotImplementedError

    def _handle_conversion(self, from_size: int, to_size: int, signed: bool, operand: pyvex.expr.IRExpr) -> Any:
        return None

    def _handle_stmt_Put(self, stmt):
        v = self._expr(stmt.data)
        # there are cases like  VMOV.F32        S0, S0
        # so we need to check if this register write is actually a no-op
        if isinstance(stmt.data, pyvex.IRExpr.RdTmp):
            t = self.state.tmps.get(stmt.data.tmp, None)
            if isinstance(t, RegisterOffset) and t.reg == stmt.offset:
                same_ins_read = False
                for i in range(self.stmt_idx, -1, -1):
                    if i >= self.block.vex.stmts_used:
                        break
                    prev_stmt = self.block.vex.statements[i]
                    if isinstance(prev_stmt, pyvex.IRStmt.IMark):
                        break
                    if isinstance(prev_stmt, pyvex.IRStmt.WrTmp) and prev_stmt.tmp == stmt.data.tmp:
                        same_ins_read = True
                        break
                if same_ins_read:
                    # we need to revert the read operation as well
                    self.state.register_read_undo(stmt.offset)
                return

        if stmt.offset == self.arch.sp_offset and isinstance(v, SpOffset):
            self.state.sp_value = v.offset
        elif stmt.offset == self.arch.bp_offset and isinstance(v, SpOffset):
            self.state.bp_value = v.offset
        else:
            self.state.register_written(stmt.offset, stmt.data.result_size(self.tyenv) // self.arch.byte_width)

    def _handle_stmt_Store(self, stmt: pyvex.IRStmt.Store):
        addr = self._expr(stmt.addr)
        if isinstance(addr, SpOffset):
            self.state.stack_written(addr.offset, stmt.data.result_size(self.tyenv) // self.arch.byte_width)
            data = self._expr(stmt.data)
            if isinstance(data, RegisterOffset) and not isinstance(data, SpOffset):
                # push reg; we record the stored register as well as the stack slot offset
                self.state.callee_stored_regs[data.reg] = u2s(addr.offset, self.arch.bits)
            if isinstance(data, SpOffset):
                self.state.simple_stack[addr.offset] = data

    def _handle_stmt_WrTmp(self, stmt: pyvex.IRStmt.WrTmp):
        v = self._expr(stmt.data)
        if v is not None:
            self.state.tmps[stmt.tmp] = v

    def _handle_expr_Const(self, expr: pyvex.IRExpr.Const):
        return expr.con.value

    def _handle_expr_GSPTR(self, expr):
        return 0

    def _handle_expr_Get(self, expr) -> SpOffset | RegisterOffset:
        if expr.offset == self.arch.sp_offset:
            return SpOffset(self.arch.bits, self.state.sp_value, is_base=False)
        if expr.offset == self.arch.bp_offset and not self.bp_as_gpr:
            return SpOffset(self.arch.bits, self.state.bp_value, is_base=False)
        bits = expr.result_size(self.tyenv)
        self.state.register_read(expr.offset, bits // self.arch.byte_width)
        return RegisterOffset(bits, expr.offset, 0)

    def _handle_expr_GetI(self, expr):  # type: ignore
        return None

    def _handle_expr_ITE(self, expr):  # type: ignore
        return None

    def _handle_expr_Load(self, expr):  # type: ignore
        addr = self._expr(expr.addr)
        if isinstance(addr, SpOffset):
            self.state.stack_read(addr.offset, expr.result_size(self.tyenv) // self.arch.byte_width)
            return self.state.simple_stack.get(addr.offset)
        return None

    def _handle_expr_RdTmp(self, expr):
        return self.state.tmps.get(expr.tmp, None)

    def _handle_expr_VECRET(self, expr):  # type: ignore
        return None

    @binop_handler
    def _handle_binop_Add(self, expr):
        op0, op1 = self._expr(expr.args[0]), self._expr(expr.args[1])
        if isinstance(op0, SpOffset) and isinstance(op1, int):
            return SpOffset(op0.bits, s2u(op0.offset + op1, op0.bits), is_base=op0.is_base)
        if isinstance(op1, SpOffset) and isinstance(op0, int):
            return SpOffset(op1.bits, s2u(op1.offset + op0, op1.bits), is_base=op1.is_base)
        return None

    @binop_handler
    def _handle_binop_Sub(self, expr):
        op0, op1 = self._expr(expr.args[0]), self._expr(expr.args[1])
        if isinstance(op0, SpOffset) and isinstance(op1, int):
            return SpOffset(op0.bits, s2u(op0.offset - op1, op0.bits), is_base=op0.is_base)
        if isinstance(op1, SpOffset) and isinstance(op0, int):
            return SpOffset(op1.bits, s2u(op1.offset - op0, op1.bits), is_base=op1.is_base)
        return None

    @binop_handler
    def _handle_binop_And(self, expr):
        op0, op1 = self._expr(expr.args[0]), self._expr(expr.args[1])
        if isinstance(op0, SpOffset):
            return op0
        if isinstance(op1, SpOffset):
            return op1
        return None


class FactCollector(Analysis):
    """
    An extremely fast analysis that extracts necessary facts of a function for CallingConventionAnalysis to make
    decision on the calling convention and prototype of a function.
    """

    def __init__(self, func: Function, max_depth: int = 100):
        self.function = func
        self._max_depth = max_depth

        self.input_args: list[SimRegArg | SimStackArg] | None = None
        self.retval_size: int | None = None

        self._analyze()

    def _analyze(self):
        # breadth-first search using function graph, collect registers and stack variables that are written to as well
        # as read from, until max_depth is reached

        end_states = self._analyze_startpoint()
        self._analyze_endpoints_for_retval_size()
        callee_restored_regs = self._analyze_endpoints_for_restored_regs()
        self._determine_input_args(end_states, callee_restored_regs)

    def _analyze_startpoint(self) -> list[FactCollectorState]:
        func_graph = self.function.transition_graph
        startpoint = self.function.startpoint
        if startpoint is None:
            return []

        bp_as_gpr = self.function.info.get("bp_as_gpr", False)
        engine = SimEngineFactCollectorVEX(self.project, bp_as_gpr)
        init_state = FactCollectorState()
        if self.project.arch.call_pushes_ret:
            init_state.sp_value = self.project.arch.bytes
        init_state.bp_value = init_state.sp_value

        traversed = set()
        queue: list[
            tuple[int, FactCollectorState, CodeNode | BlockNode | HookNode | Function, BlockNode | HookNode | None]
        ] = [(0, init_state, startpoint, None)]
        end_states: list[FactCollectorState] = []
        while queue:
            depth, state, node, retnode = queue.pop(0)
            if isinstance(node, BlockNode) and node in traversed:
                continue
            traversed.add(node)

            if depth > self._max_depth:
                end_states.append(state)
                break

            if isinstance(node, BlockNode) and node.size == 0:
                continue
            if isinstance(node, HookNode):
                # attempt to convert it into a function
                if self.kb.functions.contains_addr(node.addr):
                    node = self.kb.functions.get_by_addr(node.addr)
                else:
                    continue
            if isinstance(node, Function):
                if node.calling_convention is not None and node.prototype is not None:
                    # consume args and overwrite the return register
                    self._handle_function(state, node)
                if node.returning is False or retnode is None:
                    # the function call does not return
                    end_states.append(state)
                else:
                    # enqueue the retnode, but we don't increment the depth
                    new_state = state.copy()
                    if self.project.arch.call_pushes_ret:
                        new_state.sp_value += self.project.arch.bytes
                    queue.append((depth, new_state, retnode, None))
                continue

            block = self.project.factory.block(node.addr, size=node.size)
            engine.process(state, block=block)

            successor_added = False
            call_succ, ret_succ = None, None
            for _, succ, data in func_graph.out_edges(node, data=True):
                edge_type = data.get("type")
                outside = data.get("outside", False)
                if depth + 1 <= self._max_depth:
                    if edge_type == "fake_return":
                        if succ not in traversed:
                            ret_succ = succ
                    elif edge_type == "transition" and not outside:
                        if succ not in traversed:
                            successor_added = True
                            queue.append((depth + 1, state.copy(), succ, None))
                    elif edge_type == "call" or (edge_type == "transition" and outside):
                        # a call or a tail-call
                        # note that it's ok to traverse a called function multiple times
                        if not isinstance(succ, Function):
                            if self.kb.functions.contains_addr(succ.addr):
                                succ = self.kb.functions.get_by_addr(succ.addr)
                            else:
                                # not sure who we are calling
                                continue
                        call_succ = succ
            if call_succ is not None:
                successor_added = True
                queue.append((depth + 1, state.copy(), call_succ, ret_succ))

            if not successor_added:
                end_states.append(state)

        return end_states

    def _handle_function(self, state: FactCollectorState, func: Function) -> None:
        try:
            if func.calling_convention is not None and func.prototype is not None:
                arg_locs = func.calling_convention.arg_locs(func.prototype)
            else:
                return
        except (TypeError, ValueError):
            return

        if None in arg_locs:
            return

        for arg_loc in arg_locs:
            for loc in arg_loc.get_footprint():
                if isinstance(loc, SimRegArg):
                    state.register_read(self.project.arch.registers[loc.reg_name][0] + loc.reg_offset, loc.size)
                elif isinstance(loc, SimStackArg):
                    sp_value = state.sp_value
                    if sp_value is not None:
                        state.stack_read(sp_value + loc.stack_offset, loc.size)

        # clobber caller-saved regs
        for reg_name in func.calling_convention.CALLER_SAVED_REGS:
            offset = self.project.arch.registers[reg_name][0]
            state.register_written(offset, self.project.arch.registers[reg_name][1])

    def _analyze_endpoints_for_retval_size(self):
        """
        Analyze all endpoints to determine the return value size.
        """
        func_graph = self.function.transition_graph
        cc_cls = default_cc(
            self.project.arch.name, platform=self.project.simos.name if self.project.simos is not None else None
        )
        if cc_cls is None:
            # don't know what the calling convention may be... give up
            return
        cc = cc_cls(self.project.arch)
        if isinstance(cc.RETURN_VAL, SimRegArg):
            retreg_offset = cc.RETURN_VAL.check_offset(self.project.arch)
        else:
            return

        retval_sizes = []
        for endpoint in self.function.endpoints:
            traversed = set()
            queue: list[tuple[int, BlockNode | HookNode]] = [(0, endpoint)]
            while queue:
                depth, node = queue.pop(0)
                if isinstance(node, BlockNode) and node in traversed:
                    continue
                traversed.add(node)

                if depth > 3:
                    break

                if isinstance(node, BlockNode) and node.size == 0:
                    continue

                if isinstance(node, HookNode):
                    # attempt to convert it into a function
                    if self.kb.functions.contains_addr(node.addr):
                        node = self.kb.functions.get_by_addr(node.addr)
                    else:
                        continue
                if isinstance(node, Function):
                    if (
                        node.calling_convention is not None
                        and node.prototype is not None
                        and node.prototype.returnty is not None
                        and not isinstance(node.prototype.returnty, SimTypeBottom)
                    ):
                        # assume the function overwrites the return variable
                        returnty_size = node.prototype.returnty.with_arch(self.project.arch).size
                        assert returnty_size is not None
                        retval_size = returnty_size // self.project.arch.byte_width
                        retval_sizes.append(retval_size)
                    continue

                # if this block ends with a call to a function, we process the function first
                func_succs = [
                    succ
                    for succ in func_graph.successors(node)
                    if isinstance(succ, (Function, HookNode)) or self.kb.functions.contains_addr(succ.addr)
                ]
                if len(func_succs) == 1:
                    func_succ = func_succs[0]
                    if isinstance(func_succ, (BlockNode, HookNode)) and self.kb.functions.contains_addr(func_succ.addr):
                        # attempt to convert it into a function
                        func_succ = self.kb.functions.get_by_addr(func_succ.addr)
                    if isinstance(func_succ, Function) and func_succ.name not in {  # noqa:SIM102
                        "_security_check_cookie"
                    }:
                        if (
                            func_succ.calling_convention is not None
                            and func_succ.prototype is not None
                            and func_succ.prototype.returnty is not None
                            and not isinstance(func_succ.prototype.returnty, SimTypeBottom)
                        ):
                            # assume the function overwrites the return variable
                            proto = func_succ.prototype
                            if func_succ.prototype_libname is not None:
                                # we need to deref the prototype in case it uses SimTypeRef internally
                                proto = dereference_simtype_by_lib(proto, func_succ.prototype_libname)

                            assert isinstance(proto, SimTypeFunction) and proto.returnty is not None
                            returnty_size = proto.returnty.with_arch(self.project.arch).size
                            if returnty_size is None:
                                # it may be None if somehow we cannot resolve a SimTypeRef; we fall back to the full
                                # machine word size
                                retval_size = self.project.arch.bytes
                            else:
                                retval_size = returnty_size // self.project.arch.byte_width
                            retval_sizes.append(retval_size)
                            continue

                block = self.project.factory.block(node.addr, size=node.size)
                # scan the block statements backwards to find writes to the return value register
                retval_size = None
                for stmt in reversed(block.vex.statements):
                    if isinstance(stmt, pyvex.IRStmt.Put):
                        assert block.vex.tyenv is not None
                        size = stmt.data.result_size(block.vex.tyenv) // self.project.arch.byte_width
                        if stmt.offset == retreg_offset:
                            retval_size = max(size, 1)

                if retval_size is not None:
                    retval_sizes.append(retval_size)
                    continue

                for pred, _, data in func_graph.in_edges(node, data=True):
                    edge_type = data.get("type")
                    if pred not in traversed and depth + 1 <= self._max_depth:
                        if edge_type == "call":
                            continue
                        if edge_type in {"transition", "fake_return"}:
                            queue.append((depth + 1, pred))

        self.retval_size = max(retval_sizes) if retval_sizes else None

    def _analyze_endpoints_for_restored_regs(self):
        """
        Analyze all endpoints to determine the restored registers.
        """
        func_graph = self.function.transition_graph
        callee_restored_regs = set()

        sp_masks = {
            0xFFFFFFFE,
            0xFFFFFFFC,
            0xFFFFFFF8,
            0xFFFFFFF0,
            0xFFFFFFFF_FFFFFFFE,
            0xFFFFFFFF_FFFFFFFC,
            0xFFFFFFFF_FFFFFFF8,
            0xFFFFFFFF_FFFFFFF0,
        }
        for endpoint in self.function.endpoints:
            traversed = set()
            queue: list[tuple[int, BlockNode | HookNode]] = [(0, endpoint)]
            while queue:
                depth, node = queue.pop(0)
                traversed.add(node)

                if depth > 3:
                    break

                if isinstance(node, BlockNode) and node.size == 0:
                    continue
                if isinstance(node, (HookNode, Function)):
                    continue

                block = self.project.factory.block(node.addr, size=node.size)
                # scan the block statements backwards to find all statements that restore registers from the stack
                tmps = {}
                for stmt in block.vex.statements:
                    if isinstance(stmt, pyvex.IRStmt.WrTmp):
                        if isinstance(stmt.data, pyvex.IRExpr.Get) and stmt.data.offset in {
                            self.project.arch.bp_offset,
                            self.project.arch.sp_offset,
                        }:
                            tmps[stmt.tmp] = "sp"
                        elif (
                            isinstance(stmt.data, pyvex.IRExpr.Load)
                            and isinstance(stmt.data.addr, pyvex.IRExpr.RdTmp)
                            and tmps.get(stmt.data.addr.tmp) == "sp"
                        ):
                            tmps[stmt.tmp] = "stack_value"
                        elif isinstance(stmt.data, pyvex.IRExpr.Const):
                            tmps[stmt.tmp] = "const"
                        elif isinstance(stmt.data, pyvex.IRExpr.Binop):
                            if stmt.data.op.startswith("Iop_Add") or stmt.data.op.startswith("Iop_Sub"):
                                if (
                                    isinstance(stmt.data.args[0], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[0].tmp) == "sp"
                                ) or (
                                    isinstance(stmt.data.args[1], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[1].tmp) == "sp"
                                ):
                                    tmps[stmt.tmp] = "sp"
                            elif stmt.data.op.startswith("Iop_And"):  # noqa: SIM102
                                if (
                                    isinstance(stmt.data.args[0], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[0].tmp) == "sp"
                                    and isinstance(stmt.data.args[1], pyvex.IRExpr.Const)
                                    and stmt.data.args[1].con.value in sp_masks
                                ) or (
                                    isinstance(stmt.data.args[1], pyvex.IRExpr.RdTmp)
                                    and tmps.get(stmt.data.args[1].tmp) == "sp"
                                    and isinstance(stmt.data.args[0], pyvex.IRExpr.Const)
                                    and stmt.data.args[0].con.value in sp_masks
                                ):
                                    tmps[stmt.tmp] = "sp"
                    if isinstance(stmt, pyvex.IRStmt.Put):
                        assert block.vex.tyenv is not None
                        size = stmt.data.result_size(block.vex.tyenv) // self.project.arch.byte_width
                        # is the data loaded from the stack?
                        if (
                            size == self.project.arch.bytes
                            and isinstance(stmt.data, pyvex.IRExpr.RdTmp)
                            and tmps.get(stmt.data.tmp) == "stack_value"
                        ):
                            callee_restored_regs.add(stmt.offset)

                for pred, _, data in func_graph.in_edges(node, data=True):
                    edge_type = data.get("type")
                    if pred not in traversed and depth + 1 <= self._max_depth and edge_type == "transition":
                        queue.append((depth + 1, pred))

        # remove offsets of registers that may store return values from callee_restored_regs
        ret_reg_offsets = set()
        cc_cls = default_cc(
            self.project.arch.name, platform=self.project.simos.name if self.project.simos is not None else None
        )
        if cc_cls is not None:
            cc = cc_cls(self.project.arch)
            if isinstance(cc.RETURN_VAL, SimRegArg):
                retreg_offset = cc.RETURN_VAL.check_offset(self.project.arch)
                ret_reg_offsets.add(retreg_offset)
            if isinstance(cc.OVERFLOW_RETURN_VAL, SimRegArg):
                retreg_offset = cc.OVERFLOW_RETURN_VAL.check_offset(self.project.arch)
                ret_reg_offsets.add(retreg_offset)
            if isinstance(cc.FP_RETURN_VAL, SimRegArg):
                try:
                    retreg_offset = cc.FP_RETURN_VAL.check_offset(self.project.arch)
                    ret_reg_offsets.add(retreg_offset)
                except KeyError:
                    # register name does not exist
                    pass

        return callee_restored_regs.difference(ret_reg_offsets)

    def _determine_input_args(self, end_states: list[FactCollectorState], callee_restored_regs: set[int]) -> None:
        self.input_args = []
        reg_offset_created = set()
        callee_saved_regs = set()
        callee_saved_reg_stack_offsets = set()

        # determine callee-saved registers
        for state in end_states:
            for reg_offset, stack_offset in state.callee_stored_regs.items():
                if reg_offset in callee_restored_regs:
                    callee_saved_regs.add(reg_offset)
                    callee_saved_reg_stack_offsets.add(stack_offset)

        for state in end_states:
            for offset, size in state.reg_reads.items():
                if (
                    offset in reg_offset_created
                    or offset == self.project.arch.bp_offset
                    or not is_sane_register_variable(self.project.arch, offset, size)
                    or offset in callee_saved_regs
                ):
                    continue
                reg_offset_created.add(offset)
                reg_name = self.project.arch.translate_register_name(offset, size=size)
                arg = SimRegArg(reg_name, size)
                self.input_args.append(arg)

        stack_offset_created = set()
        ret_addr_offset = 0 if not self.project.arch.call_pushes_ret else self.project.arch.bytes
        # handle shadow stack args
        cc_cls = default_cc(
            self.project.arch.name, platform=self.project.simos.name if self.project.simos is not None else None
        )
        stackarg_sp_buff = cc_cls.STACKARG_SP_BUFF if cc_cls is not None else 0
        for state in end_states:
            for offset, size in state.stack_reads.items():
                offset = u2s(offset, self.project.arch.bits)
                if offset - ret_addr_offset > stackarg_sp_buff:
                    if offset in stack_offset_created or offset in callee_saved_reg_stack_offsets:
                        continue
                    stack_offset_created.add(offset)
                    arg = SimStackArg(offset - ret_addr_offset, size)
                    self.input_args.append(arg)


AnalysesHub.register_default("FunctionFactCollector", FactCollector)



from __future__ import annotations

from .mips_elf_fast import MipsElfFastResolver
from .mips_elf_got import MipsElfGotResolver
from .x86_elf_pic_plt import X86ElfPicPltResolver
from .jumptable import JumpTableResolver
from .x86_pe_iat import X86PeIatResolver
from .amd64_elf_got import AMD64ElfGotResolver
from .arm_elf_fast import ArmElfFastResolver
from .const_resolver import ConstantResolver
from .amd64_pe_iat import AMD64PeIatResolver
from .memload_resolver import MemoryLoadResolver
from .syscall_resolver import SyscallResolver


__all__ = (
    "AMD64ElfGotResolver",
    "AMD64PeIatResolver",
    "ArmElfFastResolver",
    "ConstantResolver",
    "JumpTableResolver",
    "MemoryLoadResolver",
    "MipsElfFastResolver",
    "MipsElfGotResolver",
    "SyscallResolver",
    "X86ElfPicPltResolver",
    "X86PeIatResolver",
)



from __future__ import annotations
from angr.ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class RemoveRedundantShiftsAroundComparators(PeepholeOptimizationExprBase):
    """
    Remove redundant bitshifts for both operands around a comparison operator.

    More cases can be added in the future as we encounter them.
    """

    __slots__ = ()

    NAME = "Remove redundant bitshifts for operands around a comparator"
    expr_classes = (BinaryOp,)  # all expressions are allowed

    def optimize(
        self, expr: BinaryOp, stmt_idx: int | None = None, block=None, **kwargs
    ):  # pylint:disable=unused-argument
        # (expr_0 << N) < (expr_1 << N)  ==> expr_0 << expr_1
        # FIXME: This optimization is unsafe but seems to work for all existing case
        if expr.op in {"CmpLE", "CmpLT", "CmpEQ", "CmpNE", "CmpGE", "CmpGT"}:
            op0, op1 = expr.operands
            if (
                isinstance(op0, BinaryOp)
                and op0.op == "Shl"
                and isinstance(op1, BinaryOp)
                and op1.op == "Shl"
                and isinstance(op0.operands[1], Const)
                and isinstance(op1.operands[1], Const)
                and op0.operands[1].value == op1.operands[1].value
            ):
                return BinaryOp(
                    expr.idx,
                    expr.op,
                    [op0.operands[0], op1.operands[0]],
                    expr.signed,
                    bits=expr.bits,
                    floating_point=expr.floating_point,
                    rounding_mode=expr.rounding_mode,
                    **expr.tags,
                )

            # might have been rewritten to multiplications
            if (
                isinstance(op0, BinaryOp)
                and op0.op == "Mul"
                and isinstance(op0.operands[1], Const)
                and op0.operands[1].is_int
            ):
                op0_op = op0.operands[0]
                mul_0 = op0.operands[1].value_int
                mul_1 = None
                op1_op = None
                if (
                    isinstance(op1, BinaryOp)
                    and op1.op == "Mul"
                    and isinstance(op1.operands[1], Const)
                    and op1.operands[1].is_int
                ):
                    op1_op = op1.operands[0]
                    mul_1 = op1.operands[1].value_int
                elif isinstance(op1, Const):
                    op1_op = None
                    mul_1 = op1.value_int

                if mul_1 is not None:
                    common_shift_amount = self._get_common_shift_amount(mul_0, mul_1)
                    if common_shift_amount > 0:
                        new_mul_0 = Const(None, None, mul_0 >> common_shift_amount, op0.bits)
                        new_mul_1 = Const(None, None, mul_1 >> common_shift_amount, op0.bits)
                        new_cmp_0 = BinaryOp(op0.idx, "Mul", [op0_op, new_mul_0], op0.signed, bits=op0.bits, **op0.tags)
                        new_cmp_1 = (
                            BinaryOp(op1.idx, "Mul", [op1_op, new_mul_1], op1.signed, bits=op1.bits, **op1.tags)
                            if op1_op is not None
                            else new_mul_1
                        )
                        return BinaryOp(
                            expr.idx,
                            expr.op,
                            [new_cmp_0, new_cmp_1],
                            expr.signed,
                            bits=expr.bits,
                            floating_point=expr.floating_point,
                            rounding_mode=expr.rounding_mode,
                            **expr.tags,
                        )

        return None

    @staticmethod
    def _get_common_shift_amount(v0: int, v1: int) -> int:
        if v0 == 0 or v1 == 0:
            return 0
        shift_amount = 0
        while (v0 & 1) == 0 and (v1 & 1) == 0:
            if v0 & 0xFFFF == 0 and v1 & 0xFFFF == 0:
                v0 >>= 16
                v1 >>= 16
                shift_amount += 16
            elif v0 & 0xFF == 0 and v1 & 0xFF == 0:
                v0 >>= 8
                v1 >>= 8
                shift_amount += 8
            elif v0 & 0xF == 0 and v1 & 0xF == 0:
                v0 >>= 4
                v1 >>= 4
                shift_amount += 4
            else:
                v0 >>= 1
                v1 >>= 1
                shift_amount += 1
        return shift_amount



from __future__ import annotations
import importlib.util
import sys


def lazy_import(name):
    spec = importlib.util.find_spec(name)
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module



from __future__ import annotations

try:
    import sqlalchemy
except ImportError as err:
    sqlalchemy = None
    raise ImportError(
        "AngrDB relies on SQLAlchemy. Please install SQLAlchemy first by running:\n\tpip install sqlalchemy"
    ) from err

from .db import AngrDB


__all__ = ("AngrDB",)



from __future__ import annotations
from typing import Any

from enum import Enum


class DecompilationNoteLevel(Enum):
    """
    Enum class describing the level of each decompilation note.
    """

    DEBUG = 0
    INFO = 1
    WARNING = 2
    CRITICAL = 3


class DecompilationNote:
    """
    Describes a note that is generated during decompilation.

    Key is a unique string for the decompilation note. It is used as an index in the decompilation notes dictionary in
    the Decompiler class.
    Name is string for display by default.
    Content is the actual content of the note. It can be of any time, but for custom types, you must override `__str__`
    so that it can be displayed.
    Level is the level of the note. The following values are available: DecompilationNoteLevel.DEBUG,
    DecompilationNoteLevel.INFO, DecompilationNoteLevel.WARNING, and DecompilationNoteLevel.CRITICAL.
    """

    __slots__ = (
        "content",
        "key",
        "level",
        "name",
    )

    def __init__(self, key: str, name: str, content: Any, *, level=DecompilationNoteLevel.INFO):
        self.key = key
        self.name = name
        self.content = content
        self.level = level

    def __repr__(self):
        return f"<DecompilationNote: {self.name}>"

    def __str__(self):
        return f"{self.name}: {self.content}"



# pylint:disable=raise-missing-from
from __future__ import annotations

from typing import TypeVar, Generic, cast, TYPE_CHECKING, overload
from collections.abc import Iterator
import contextlib
from collections.abc import Generator
import logging
import collections.abc
import re
import weakref
import bisect
import os

import networkx

from archinfo.arch_soot import SootMethodDescriptor
import cle

from angr.errors import SimEngineError
from angr.knowledge_plugins.plugin import KnowledgeBasePlugin
from .function import Function
from .soot_function import SootFunction

K = TypeVar("K", int, SootMethodDescriptor)
T = TypeVar("T")

if TYPE_CHECKING:
    from angr import KnowledgeBase

    class SortedDict(Generic[K, T], dict[K, T]):
        def irange(self, *args, **kwargs) -> Iterator[K]: ...

else:
    from sortedcontainers import SortedDict


QUERY_PATTERN = re.compile(r"^(::(.+?))?::(.+)$")
ADDR_PATTERN = re.compile(r"^(0x[\dA-Fa-f]+)|(\d+)$")

l = logging.getLogger(name=__name__)
_missing = object()


class FunctionDict(Generic[K], SortedDict[K, Function]):
    """
    FunctionDict is a dict where the keys are function starting addresses and
    map to the associated :class:`Function`.
    """

    def __init__(self, backref: FunctionManager[K] | None, *args, key_types: type = int, **kwargs):
        self._backref = (
            cast(FunctionManager[K], backref if isinstance(backref, weakref.ProxyType) else weakref.proxy(backref))
            if backref is not None
            else None
        )
        self._key_types = key_types
        super().__init__(*args, **kwargs)

    def copy(self) -> FunctionDict[K]:
        return FunctionDict(self._backref, self, key_types=self._key_types)

    def __getitem__(self, addr: K) -> Function:
        try:
            return super().__getitem__(addr)
        except KeyError as ex:
            if isinstance(addr, bool) or not isinstance(addr, self._key_types):
                raise TypeError(f"FunctionDict only supports {self._key_types} as key type") from ex

            if isinstance(addr, SootMethodDescriptor):
                t = SootFunction(self._backref, addr)
            else:
                t = Function(self._backref, addr)
            with contextlib.suppress(Exception):
                self[addr] = t
            if self._backref is not None:
                self._backref._function_added(t)
            return t

    @overload
    def get(self, key: K, default: None = None, /) -> Function: ...
    @overload
    def get(self, key: K, default: Function, /) -> Function: ...
    @overload
    def get(self, key: K, default: T, /) -> Function | T: ...

    def get(self, addr, default=_missing, /):
        try:
            return super().__getitem__(addr)
        except KeyError:
            if default is _missing:
                raise
            return default

    def floor_addr(self, addr):
        try:
            return next(self.irange(maximum=addr, reverse=True))
        except StopIteration as err:
            raise KeyError(addr) from err

    def ceiling_addr(self, addr):
        try:
            return next(self.irange(minimum=addr, reverse=False))
        except StopIteration as err:
            raise KeyError(addr) from err

    def __setstate__(self, state):
        for v, k in state.items():
            self[k] = v

    def __getstate__(self):
        return dict(self.items())


class FunctionManager(Generic[K], KnowledgeBasePlugin, collections.abc.Mapping[K, Function]):
    """
    This is a function boundaries management tool. It takes in intermediate
    results during CFG generation, and manages a function map of the binary.
    """

    def __init__(self, kb: KnowledgeBase):
        super().__init__(kb=kb)
        self.function_address_types = self._kb._project.arch.function_address_types
        self.address_types = self._kb._project.arch.address_types
        self._function_map: FunctionDict[K] = FunctionDict(self, key_types=self.function_address_types)
        self.function_addrs_set: set = set()
        self.callgraph = networkx.MultiDiGraph()
        self.block_map = {}

        # Registers used for passing arguments around
        self._arg_registers = kb._project.arch.argument_registers

        # local PLT dictionary cache
        self._rplt_cache_ranges: None | list[tuple[int, int]] = None
        self._rplt_cache: None | set[int] = None
        # local binary name cache: min_addr -> (max_addr, binary_name)
        self._binname_cache: None | SortedDict[int, tuple[int, str | None]] = None

    def __setstate__(self, state):
        self._kb = state["_kb"]
        self.function_address_types = state["function_address_types"]
        self.address_types = state["address_types"]
        self._function_map = state["_function_map"]
        self.callgraph = state["callgraph"]
        self.block_map = state["block_map"]

        self._function_map._backref = weakref.proxy(self)
        for func in self._function_map.values():
            func._function_manager = self

    def __getstate__(self):
        return {
            "_kb": self._kb,
            "function_address_types": self.function_address_types,
            "address_types": self.address_types,
            "_function_map": self._function_map,
            "callgraph": self.callgraph,
            "block_map": self.block_map,
        }

    def copy(self):
        fm = FunctionManager(self._kb)
        fm._function_map = self._function_map.copy()
        for address, function in fm._function_map.items():
            fm._function_map[address] = function.copy()
        fm.callgraph = networkx.MultiDiGraph(self.callgraph)
        fm._arg_registers = self._arg_registers.copy()
        fm.function_addrs_set = self.function_addrs_set.copy()

        return fm

    def clear(self):
        self._function_map = FunctionDict(self, key_types=self.function_address_types)
        self.callgraph = networkx.MultiDiGraph()
        self.block_map.clear()
        self.function_addrs_set = set()
        # cache
        self._rplt_cache = None
        self._rplt_cache_ranges = None
        self._binname_cache = None

    def _genenate_callmap_sif(self, filepath):
        """
        Generate a sif file from the call map.

        :param filepath:    Path of the sif file
        :return:            None
        """
        with open(filepath, "w", encoding="utf-8") as f:
            f.writelines(f"{src:#x}\tDirectEdge\t{dst:#x}\n" for src, dst in self.callgraph.edges())

    def _addr_in_plt_cached_ranges(self, addr: int) -> bool:
        if self._rplt_cache_ranges is None:
            return False
        pos = bisect.bisect_left(self._rplt_cache_ranges, addr, key=lambda x: x[0])
        return pos > 0 and self._rplt_cache_ranges[pos - 1][0] <= addr < self._rplt_cache_ranges[pos - 1][1]

    def is_plt_cached(self, addr: int) -> bool:
        # check if the addr is in the cache range
        if not self._addr_in_plt_cached_ranges(addr):
            # find the object containing this addr
            obj = self._kb._project.loader.find_object_containing(addr, membership_check=False)
            if obj is None:
                return False
            if self._rplt_cache_ranges is None:
                self._rplt_cache_ranges = []
            obj_range = obj.min_addr, obj.max_addr
            idx = bisect.bisect_left(self._rplt_cache_ranges, obj_range)
            if not (idx < len(self._rplt_cache_ranges) and self._rplt_cache_ranges[idx] == obj_range):
                self._rplt_cache_ranges.insert(idx, obj_range)
            if isinstance(obj, (cle.MetaELF, cle.MachO)):
                if self._rplt_cache is None:
                    self._rplt_cache = set()
                self._rplt_cache |= set(obj.reverse_plt)

        return addr in self._rplt_cache if self._rplt_cache is not None else False

    def _binname_cache_get_addr_base(self, addr: int) -> int | None:
        if self._binname_cache is None:
            return None
        try:
            base_addr = next(self._binname_cache.irange(maximum=addr, reverse=True))
        except StopIteration:
            return None
        return base_addr if base_addr <= addr < self._binname_cache[base_addr][0] else None

    def get_binary_name_cached(self, addr: int) -> str | None:
        base_addr = self._binname_cache_get_addr_base(addr)
        if base_addr is None:
            # not cached; cache it first
            obj = self._kb._project.loader.find_object_containing(addr, membership_check=False)
            if obj is None:
                return None
            if self._binname_cache is None:
                self._binname_cache = SortedDict()
            binary_basename = os.path.basename(obj.binary) if obj.binary else None
            self._binname_cache[obj.min_addr] = obj.max_addr, binary_basename
            base_addr = obj.min_addr
        return self._binname_cache[base_addr][1] if self._binname_cache is not None else None

    def _add_node(self, function_addr, node, syscall=None, size=None):
        if isinstance(node, self.address_types):
            node = self._kb._project.factory.snippet(node, size=size)
        dst_func = self._function_map[function_addr]
        if syscall in (True, False):
            dst_func.is_syscall = syscall
        dst_func._register_node(True, node)
        self.block_map[node.addr] = node

    def _add_call_to(
        self,
        function_addr,
        from_node,
        to_addr,
        retn_node=None,
        syscall=None,
        stmt_idx=None,
        ins_addr=None,
        return_to_outside: bool = False,
    ):
        """
        Add a call to a function.

        :param int function_addr:   Address of the current function where this call happens.
        :param from_node:           The source node.
        :param to_addr:             Address of the target function, or None if unknown.
        :param retn_node:           The node where the target function will return to if it returns.
        :param bool syscall:        If this is a call to a syscall or not.
        :param int stmt_idx:        ID of the statement where this call happens.
        :param int ins_addr:        Address of the instruction where this call happens.
        :param return_to_outside:  True if the return of the call is considered going to outside of the current
                                        function.
        :return:                    None
        """

        if isinstance(from_node, self.address_types):
            from_node = self._kb._project.factory.snippet(from_node)
        if isinstance(retn_node, self.address_types):
            retn_node = self._kb._project.factory.snippet(retn_node)
        func = self._function_map[function_addr]
        func._add_call_site(from_node.addr, to_addr, retn_node.addr if retn_node else None)

        if to_addr is not None:
            dest_func = self._function_map[to_addr]
            if syscall in (True, False):
                dest_func.is_syscall = syscall
            func._call_to(
                from_node,
                dest_func,
                retn_node,
                stmt_idx=stmt_idx,
                ins_addr=ins_addr,
                return_to_outside=return_to_outside,
            )

        if return_to_outside:
            func.add_retout_site(from_node)

        # is there any existing edge on the callgraph?
        edge_data = {"type": "call"}
        if to_addr is not None and (
            function_addr not in self.callgraph
            or to_addr not in self.callgraph[function_addr]
            or edge_data not in self.callgraph[function_addr][to_addr].values()
        ):
            self.callgraph.add_edge(function_addr, to_addr, **edge_data)

    def _add_fakeret_to(
        self, function_addr, from_node, to_node, confirmed=None, syscall=None, to_outside=False, to_function_addr=None
    ):
        if isinstance(from_node, self.address_types):
            from_node = self._kb._project.factory.snippet(from_node)
        if isinstance(to_node, self.address_types):
            to_node = self._kb._project.factory.snippet(to_node)
        src_func = self._function_map[function_addr]

        if syscall in (True, False):
            src_func.is_syscall = syscall

        src_func._fakeret_to(from_node, to_node, confirmed=confirmed, to_outside=to_outside)

        if to_outside and to_function_addr is not None:
            # mark it on the callgraph
            edge_data = {"type": "fakeret"}
            if (
                function_addr not in self.callgraph
                or to_function_addr not in self.callgraph[function_addr]
                or edge_data not in self.callgraph[function_addr][to_function_addr].values()
            ):
                self.callgraph.add_edge(function_addr, to_function_addr, **edge_data)

    def _remove_fakeret(self, function_addr, from_node, to_node):
        if type(from_node) is int:  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        if type(to_node) is int:  # pylint: disable=unidiomatic-typecheck
            to_node = self._kb._project.factory.snippet(to_node)
        self._function_map[function_addr]._remove_fakeret(from_node, to_node)

    def _add_return_from(self, function_addr, from_node, to_node=None):  # pylint:disable=unused-argument
        if isinstance(from_node, self.address_types):  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        self._function_map[function_addr]._add_return_site(from_node)

    def _add_transition_to(self, function_addr, from_node, to_node, ins_addr=None, stmt_idx=None, is_exception=False):
        if isinstance(from_node, self.address_types):  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        if isinstance(to_node, self.address_types):  # pylint: disable=unidiomatic-typecheck
            to_node = self._kb._project.factory.snippet(to_node)
        self._function_map[function_addr]._transit_to(
            from_node, to_node, ins_addr=ins_addr, stmt_idx=stmt_idx, is_exception=is_exception
        )

    def _add_outside_transition_to(
        self, function_addr, from_node, to_node, to_function_addr=None, ins_addr=None, stmt_idx=None, is_exception=False
    ):
        if type(from_node) is int:  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        if type(to_node) is int:  # pylint: disable=unidiomatic-typecheck
            try:
                to_node = self._kb._project.factory.snippet(to_node)
            except SimEngineError:
                # we cannot get the snippet, but we should at least tell the function that it's going to jump out here
                self._function_map[function_addr].add_jumpout_site(from_node)
                return
        self._function_map[function_addr]._transit_to(
            from_node,
            to_node,
            outside=True,
            ins_addr=ins_addr,
            stmt_idx=stmt_idx,
            is_exception=is_exception,
        )

        if to_function_addr is not None:
            # mark it on the callgraph
            edge_data = {"type": "transition" if not is_exception else "exception"}
            if (
                function_addr not in self.callgraph
                or to_function_addr not in self.callgraph[function_addr]
                or edge_data not in self.callgraph[function_addr][to_function_addr].values()
            ):
                self.callgraph.add_edge(function_addr, to_function_addr, **edge_data)

    def _add_return_from_call(self, function_addr, src_function_addr, to_node, to_outside=False):
        # Note that you will never return to a syscall

        if type(to_node) is int:  # pylint: disable=unidiomatic-typecheck
            to_node = self._kb._project.factory.snippet(to_node)
        func = self._function_map[function_addr]
        src_func = self._function_map[src_function_addr]
        func._return_from_call(src_func, to_node, to_outside=to_outside)

    #
    # Dict methods
    #

    def __contains__(self, item):
        if type(item) is int:
            # this is an address
            return item in self._function_map

        try:
            _ = self[item]
            return True
        except (KeyError, TypeError):
            return False

    def __getitem__(self, k) -> Function:
        if isinstance(k, self.function_address_types):
            f = self.function(addr=k)
        elif type(k) is str:
            f = self.function(name=k) or self.function(name=k, check_previous_names=True)
        else:
            raise ValueError(f"FunctionManager.__getitem__ does not support keys of type {type(k)}")

        if f is None:
            raise KeyError(k)

        return f

    def __setitem__(self, k, v):
        if isinstance(k, self.function_address_types):
            self._function_map[k] = v
            self._function_added(v)
        else:
            raise ValueError("FunctionManager.__setitem__ keys must be an int")

    def __delitem__(self, k):
        if isinstance(k, self.function_address_types):
            del self._function_map[k]
            if k in self.callgraph:
                self.callgraph.remove_node(k)
            self.function_addrs_set.discard(k)
        else:
            raise ValueError(
                f"FunctionManager.__delitem__ only accepts the following address types: "
                f"{self.function_address_types}"
            )

    def __len__(self):
        return len(self._function_map)

    def __iter__(self):
        yield from sorted(self._function_map.keys())

    def get_by_addr(self, addr) -> Function:
        return self._function_map.get(addr)

    def get_by_name(self, name: str, check_previous_names: bool = False) -> Generator[Function]:
        for f in self._function_map.values():
            if f.name == name or (check_previous_names and name in f.previous_names):
                yield f

    def _function_added(self, func: Function):
        """
        A callback method for adding a new function instance to the manager.

        :param func:   The Function instance being added.
        :return:       None
        """

        # Add the function address to the set of function addresses
        self.function_addrs_set.add(func.addr)

        # make sure all functions exist in the call graph
        self.callgraph.add_node(func.addr)

    def contains_addr(self, addr):
        """
        Decide if an address is handled by the function manager.

        Note: this function is non-conformant with python programming idioms, but its needed for performance reasons.

        :param int addr: Address of the function.
        """
        return addr in self._function_map

    def ceiling_func(self, addr):
        """
        Return the function who has the least address that is greater than or equal to `addr`.

        :param int addr: The address to query.
        :return:         A Function instance, or None if there is no other function after `addr`.
        :rtype:          Function or None
        """

        try:
            next_addr = self._function_map.ceiling_addr(addr)
            return self._function_map.get(next_addr)

        except KeyError:
            return None

    def floor_func(self, addr):
        """
        Return the function who has the greatest address that is less than or equal to `addr`.

        :param int addr: The address to query.
        :return:         A Function instance, or None if there is no other function before `addr`.
        :rtype:          Function or None
        """

        try:
            prev_addr = self._function_map.floor_addr(addr)
            return self._function_map.get(prev_addr)

        except KeyError:
            return None

    def query(self, query: str, check_previous_names: bool = False) -> Function | None:
        """
        Query for a function using selectors to disambiguate. Supported variations:

            ::<name>           Function <name> in the main object
            ::<addr>::<name>   Function <name> at <addr>
            ::<obj>::<name>    Function <name> in <obj>

        """
        # FIXME: Proper mangle handling
        matches = QUERY_PATTERN.match(query)
        if matches:
            selector = matches.group(2)
            name = matches.group(3)

            if selector is not None and ADDR_PATTERN.fullmatch(selector):
                addr = cast(K, int(matches.group(2), 0))
                try:
                    func = self._function_map.get(addr)
                    if func.name == name or (check_previous_names and name in func.previous_names):
                        return func
                except KeyError:
                    pass

            obj_name = selector or self._kb._project.loader.main_object.binary_basename
            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                if func.binary_name == obj_name:
                    return func

        return None

    def function(
        self, addr=None, name=None, check_previous_names=False, create=False, syscall=False, plt=None
    ) -> Function | None:
        """
        Get a function object from the function manager.

        Pass either `addr` or `name` with the appropriate values.

        :param int addr: Address of the function.
        :param str name: Name of the function.
        :param bool create: Whether to create the function or not if the function does not exist.
        :param bool syscall: True to create the function as a syscall, False otherwise.
        :param bool or None plt: True to find the PLT stub, False to find a non-PLT stub, None to disable this
                                 restriction.
        :return: The Function instance, or None if the function is not found and create is False.
        :rtype: Function or None
        """
        if name is not None and name.startswith("sub_"):
            # first check if a function with the specified name exists
            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                if plt is None or func.is_plt == plt:
                    return func

            # then enter the syntactic sugar mode
            try:
                addr = cast(K, int(name.split("_")[-1], 16))
                name = None
            except ValueError:
                pass

        if addr is not None:
            try:
                f = self._function_map.get(addr)
                if plt is None or f.is_plt == plt:
                    return f
            except KeyError:
                if create:
                    # the function is not found
                    f = self._function_map[addr]
                    if name is not None:
                        f.name = name
                    if syscall:
                        f.is_syscall = True
                    return f
        elif name is not None:
            func = self.query(name, check_previous_names=check_previous_names)
            if func is not None:
                return func

            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                if plt is None or func.is_plt == plt:
                    return func

        return None

    def dbg_draw(self, prefix="dbg_function_"):
        for func_addr, func in self._function_map.items():
            filename = f"{prefix}{func_addr:#08x}.png"
            func.dbg_draw(filename)

    def rebuild_callgraph(self):
        self.callgraph = networkx.MultiDiGraph()
        cfg = self._kb.cfgs.get_most_accurate()
        for func_addr in self._function_map:
            self.callgraph.add_node(func_addr)
        for func in self._function_map.values():
            if func.block_addrs_set:
                for node in func.transition_graph.nodes():
                    if isinstance(node, Function):
                        self.callgraph.add_edge(func.addr, node.addr)
                    else:
                        cfgnode = cfg.get_any_node(node.addr)
                        if (
                            cfgnode is not None
                            and cfgnode.function_address is not None
                            and cfgnode.function_address != func.addr
                        ):
                            self.callgraph.add_edge(func.addr, cfgnode.function_address)


KnowledgeBasePlugin.register_default("functions", FunctionManager)



# pylint:disable=arguments-differ,too-many-boolean-expressions
from __future__ import annotations

from angr.ailment import Const
from angr.ailment.expression import BinaryOp, Load, Expression, Tmp
from angr.ailment.statement import CAS, ConditionalJump, Statement, Assignment, Call

from .base import PeepholeOptimizationMultiStmtBase


_INTRINSICS_NAMES = {
    "xchg8": {"Win32": "InterlockedExchange8", "Linux": "atomic_exchange"},
    "xchg16": {"Win32": "InterlockedExchange16", "Linux": "atomic_exchange"},
    "xchg32": {"Win32": "InterlockedExchange", "Linux": "atomic_exchange"},
    "xchg64": {"Win32": "InterlockedExchange64", "Linux": "atomic_exchange"},
    "cmpxchg16": {"Win32": "InterlockedCompareExchange16", "Linux": "atomic_compare_exchange"},
    "cmpxchg32": {"Win32": "InterlockedCompareExchange", "Linux": "atomic_compare_exchange"},
    "cmpxchg64": {"Win32": "InterlockedCompareExchange64", "Linux": "atomic_compare_exchange"},
    "cmpxchg128": {"Win32": "InterlockedCompareExchange128", "Linux": "atomic_compare_exchange"},
    "lock_inc16": {"Win32": "InterlockedIncrement16", "Linux": "atomic_fetch_add"},
    "lock_inc32": {"Win32": "InterlockedIncrement", "Linux": "atomic_fetch_add"},
    "lock_inc64": {"Win32": "InterlockedIncrement64", "Linux": "atomic_fetch_add"},
    "lock_dec16": {"Win32": "InterlockedDecrement16", "Linux": "atomic_fetch_dec"},
    "lock_dec32": {"Win32": "InterlockedDecrement", "Linux": "atomic_fetch_dec"},
    "lock_dec64": {"Win32": "InterlockedDecrement64", "Linux": "atomic_fetch_dec"},
    "lock_xadd32": {"Win32": "InterlockedExchangeAdd", "Linux": "atomic_exchange_add"},
    "lock_xadd64": {"Win32": "InterlockedExchangeAdd64", "Linux": "atomic_exchange_add"},
}


class CASIntrinsics(PeepholeOptimizationMultiStmtBase):
    """
    Rewrite lock-prefixed instructions (or rather, their VEX/AIL forms) into intrinsic calls.

    Case 1.

                 mov eax, r12d
    0x140014b57: xchg eax, [0x14000365f8]

    LABEL_0x140014b57:
    CAS(0x1400365f8<64>, Conv(64->32, vvar_365{reg 112}), Load(addr=0x1400365f8<64>, size=4, endness=Iend_LE),
        vvar_27756)
    if (CasCmpNE(vvar_27756, g_1400365f8))
        goto LABEL_0x140014b57;

    => vvar_27756 = _InterlockedExchange(0x1400365f8, vvar_365{reg 112})


    Case 2.

    lock cmpxchg cs:g_WarbirdSecureFunctionsLock, r14d

    CAS(0x1400365f8<64>, 0x1<32>, 0x0<32>, vvar_27751)

    => var_27751 = _InterlockedCompareExchange(0x1400365f8, 0x1<32>, 0x0<32>)
    """

    __slots__ = ()

    NAME = "Rewrite compare-and-swap instructions into intrinsics."
    stmt_classes = ((CAS, ConditionalJump), (CAS, Statement))

    def optimize(self, stmts: list[Statement], stmt_idx: int | None = None, block=None, **kwargs):
        assert len(stmts) == 2
        cas_stmt = stmts[0]
        next_stmt = stmts[1]
        assert isinstance(cas_stmt, CAS)

        # TODO: We ignored endianness. Are there cases where the endianness is different from the host's?

        if (
            isinstance(next_stmt, ConditionalJump)
            and isinstance(next_stmt.condition, BinaryOp)
            and next_stmt.condition.op == "CasCmpNE"
            and next_stmt.tags["ins_addr"] == cas_stmt.tags["ins_addr"]
        ):
            addr = cas_stmt.addr
            expd_lo = self._resolve_tmp_expr(cas_stmt.expd_lo, block)
            next_stmt_cond_op1 = self._resolve_tmp_expr(next_stmt.condition.operands[1], block)
            if (
                isinstance(expd_lo, Load)
                and expd_lo.addr.likes(addr)
                and isinstance(next_stmt_cond_op1, Load)
                and next_stmt_cond_op1.addr.likes(addr)
                and cas_stmt.old_lo.likes(next_stmt.condition.operands[0])
                and cas_stmt.old_hi is None
            ):
                # TODO: Support cases where cas_stmt.old_hi is not None
                # Case 1

                call_expr = None
                if isinstance(cas_stmt.data_lo, BinaryOp):
                    if cas_stmt.data_lo.op == "Add" and cas_stmt.data_lo.operands[0].likes(cas_stmt.expd_lo):
                        if isinstance(cas_stmt.data_lo.operands[1], Const) and cas_stmt.data_lo.operands[1].value == 1:
                            # lock inc
                            call_expr = Call(
                                cas_stmt.idx,
                                self._get_instrincs_name(f"lock_inc{cas_stmt.bits}"),
                                args=[cas_stmt.addr],
                                bits=cas_stmt.bits,
                                ins_addr=cas_stmt.tags["ins_addr"],
                            )
                        else:
                            # lock xadd
                            call_expr = Call(
                                cas_stmt.idx,
                                self._get_instrincs_name(f"lock_xadd{cas_stmt.bits}"),
                                args=[cas_stmt.addr, cas_stmt.data_lo.operands[1]],
                                bits=cas_stmt.bits,
                                ins_addr=cas_stmt.tags["ins_addr"],
                            )
                    elif (
                        cas_stmt.data_lo.op == "Sub"
                        and cas_stmt.data_lo.operands[0].likes(cas_stmt.expd_lo)
                        and isinstance(cas_stmt.data_lo.operands[1], Const)
                        and cas_stmt.data_lo.operands[1].value == 1
                    ):
                        # lock dec
                        call_expr = Call(
                            cas_stmt.idx,
                            self._get_instrincs_name(f"lock_dec{cas_stmt.bits}"),
                            args=[cas_stmt.addr],
                            bits=cas_stmt.bits,
                            ins_addr=cas_stmt.tags["ins_addr"],
                        )

                if call_expr is None:
                    call_expr = Call(
                        cas_stmt.idx,
                        self._get_instrincs_name(f"xchg{cas_stmt.bits}"),
                        args=[addr, cas_stmt.data_lo],
                        bits=cas_stmt.bits,
                        ins_addr=cas_stmt.tags["ins_addr"],
                    )

                assignment_dst = cas_stmt.old_lo
                stmt = Assignment(cas_stmt.idx, assignment_dst, call_expr, **cas_stmt.tags)  # type:ignore
                return [stmt]

        if next_stmt.tags["ins_addr"] <= cas_stmt.tags["ins_addr"]:
            # avoid matching against statements prematurely
            return None

        if cas_stmt.old_hi is None:
            # TODO: Support cases where cas_stmt.old_hi is not None
            # Case 2
            call_expr = Call(
                cas_stmt.idx,
                self._get_instrincs_name(f"cmpxchg{cas_stmt.bits}"),
                args=[
                    cas_stmt.addr,
                    cas_stmt.data_lo,
                    cas_stmt.expd_lo,
                ],
                bits=cas_stmt.bits,
                ins_addr=cas_stmt.tags["ins_addr"],
            )
            assignment_dst = cas_stmt.old_lo
            stmt = Assignment(cas_stmt.idx, assignment_dst, call_expr, **cas_stmt.tags)  # type:ignore
            return [stmt, next_stmt]

        return None

    def _get_instrincs_name(self, mnemonic: str) -> str:
        if mnemonic in _INTRINSICS_NAMES:
            os = (
                self.project.simos.name
                if self.project is not None and self.project.simos is not None and self.project.simos.name is not None
                else "Linux"
            )
            if os not in _INTRINSICS_NAMES[mnemonic]:
                os = "Linux"
            return _INTRINSICS_NAMES[mnemonic][os]
        return mnemonic

    @staticmethod
    def _resolve_tmp_expr(expr: Expression, block) -> Expression:
        if isinstance(expr, Tmp):
            for stmt in block.statements:
                if isinstance(stmt, Assignment) and stmt.dst.likes(expr):
                    return stmt.src
        return expr



from __future__ import annotations
import angr


class closelog(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self):
        # A stub for closelog that does not do anything yet.
        return



from __future__ import annotations

from .light import VEXMixin
from .resilience import VEXResilienceMixin
from .slicing import VEXSlicingMixin

__all__ = (
    "VEXMixin",
    "VEXResilienceMixin",
    "VEXSlicingMixin",
)



# pylint:disable=no-self-use,unused-argument,too-many-boolean-expressions
from __future__ import annotations
from typing import TYPE_CHECKING

from angr.analyses import Analysis, register_analysis
from angr.analyses.decompiler.structured_codegen.c import CStructuredCodeWalker, CVariable, CConstant, CBinaryOp

if TYPE_CHECKING:
    from angr.analyses.decompiler.structured_codegen.c import (
        CAssignment,
        CFunction,
        CWhileLoop,
        CDoWhileLoop,
        CForLoop,
        CBreak,
        CStatements,
    )


class ASTNodeBase:
    """
    The base node for AST nodes used in loop analysis.
    """


class VarNode(ASTNodeBase):
    """
    Represents a variable node in the AST; corresponds to a virtual variable in AIL.
    """

    def __init__(self, var_ident: str):
        self.var_ident = var_ident

    def __repr__(self):
        return self.var_ident

    def __eq__(self, other):
        if not isinstance(other, VarNode):
            return False
        return self.var_ident == other.var_ident


class ConstNode(ASTNodeBase):
    """
    Represents a constant value.
    """

    def __init__(self, value: int):
        self.value = value

    def __repr__(self):
        return str(self.value)

    def __eq__(self, other):
        if not isinstance(other, ConstNode):
            return False
        return self.value == other.value


class BinOpNode(ASTNodeBase):
    """
    Represents a binary operation between a variable and a constant.
    """

    def __init__(self, op: str, lhs: VarNode, rhs: ConstNode):
        self.op = op
        self.lhs = lhs
        self.rhs = rhs

    def __repr__(self):
        return f"({self.lhs} {self.op} {self.rhs.value})"

    def __eq__(self, other):
        if not isinstance(other, BinOpNode):
            return False
        return self.op == other.op and self.lhs == other.lhs and self.rhs == other.rhs


class AssignmentNode(ASTNodeBase):
    """
    Represents an assignment operation in the AST.
    """

    def __init__(self, lhs: VarNode, rhs: ASTNodeBase):
        self.lhs = lhs
        self.rhs = rhs

    def __repr__(self):
        return f"{self.lhs} = {self.rhs}"

    def __eq__(self, other):
        if not isinstance(other, AssignmentNode):
            return False
        return self.lhs == other.lhs and self.rhs == other.rhs


class AssignmentToASTVisitor(CStructuredCodeWalker):
    """
    A visitor that converts AIL assignments into AST nodes.
    """

    def handle_CAssignment(self, obj: CAssignment) -> AssignmentNode | None:
        lhs = self.handle(obj.lhs)
        if lhs is None:
            return None
        rhs = self.handle(obj.rhs)
        if rhs is None:
            return None
        return AssignmentNode(lhs, rhs)

    def handle_CVariable(self, obj: CVariable) -> VarNode | None:
        if obj.unified_variable is None:
            return None
        return VarNode(obj.unified_variable.ident)

    def handle_CBinaryOp(self, obj: CBinaryOp) -> BinOpNode | None:
        match obj.op:
            case "Add":
                op = "+"
            case "Sub":
                op = "-"
            case "CmpEQ":
                op = "=="
            case "CmpNE":
                op = "!="
            case "CmpLT":
                op = "<"
            case "CmpLE":
                op = "<="
            case "CmpGT":
                op = ">"
            case "CmpGE":
                op = ">="
            case _:
                return None

        left = self.handle(obj.lhs)
        right = self.handle(obj.rhs)
        if not isinstance(left, VarNode) or not isinstance(right, ConstNode):
            return None

        return BinOpNode(op, left, right)

    def handle_CConstant(self, obj: CConstant) -> ConstNode | None:
        if isinstance(obj.value, int):
            return ConstNode(obj.value)
        return None


class LoopBodyControlStatementCollector(CStructuredCodeWalker):
    """
    A visitor that determines if there are control statements (break, continue, goto) within a loop body.
    """

    def __init__(self):
        super().__init__()
        self.has_break = False
        self.has_continue = False
        self.has_goto = False

    def handle_CBreak(self, obj: CBreak):
        self.has_break = True

    def handle_CContinue(self, obj):
        self.has_continue = True

    def handle_CGoto(self, obj):
        self.has_goto = True

    def handle_CSwitchCase(self, obj):
        return

    def handle_CForLoop(self, obj):
        return

    def handle_CDoWhileLoop(self, obj):
        return

    def handle_CWhileLoop(self, obj):
        return


class LoopBodyAssignmentCollector(CStructuredCodeWalker):
    """
    A visitor that collects assignments to specific variables within a loop body.
    """

    def __init__(self, var_idents: set[str]):
        super().__init__()

        self.var_idents = var_idents
        self.assignments: list[AssignmentNode] = []

    def handle_CAssignment(self, obj: CAssignment):
        include = False

        if isinstance(obj.lhs, CVariable) and obj.lhs.unified_variable.ident in self.var_idents:
            include = True
        elif isinstance(obj.rhs, CVariable):
            if obj.rhs.unified_variable.ident in self.var_idents:
                include = True
                if isinstance(obj.lhs, CVariable) and obj.lhs.unified_variable is not None:
                    self.var_idents.add(obj.lhs.unified_variable.ident)
        elif isinstance(obj.rhs, CBinaryOp) and (
            (isinstance(obj.rhs.lhs, CVariable) and obj.rhs.lhs.unified_variable.ident in self.var_idents)
            or (isinstance(obj.rhs.rhs, CVariable) and obj.rhs.rhs.unified_variable.ident in self.var_idents)
        ):
            include = True
            if isinstance(obj.lhs, CVariable) and obj.lhs.unified_variable is not None:
                self.var_idents.add(obj.lhs.unified_variable.ident)

        if include:
            ast_ = AssignmentToASTVisitor().handle(obj)
            if ast_ is not None:
                self.assignments.append(ast_)

    def handle_CIfElse(self, obj):
        return

    def handle_CSwitchCase(self, obj):
        return

    def handle_CForLoop(self, obj):
        return

    def handle_CDoWhileLoop(self, obj):
        return

    def handle_CWhileLoop(self, obj):
        return


class LoopVisitor(CStructuredCodeWalker):
    """
    A visitor that analyzes loop structures in CStructuredCode and collects relevant information for loops.
    """

    def __init__(self):
        super().__init__()

        self.var_values: dict[str, int | None] = {}
        self.result: dict[str, dict] = {}
        self._block_addrs_stack: list[set[int]] = []

    def _enter_loop(self):
        self._block_addrs_stack.append(set())

    def _leave_loop(self):
        if self._block_addrs_stack:
            self._block_addrs_stack.pop()

    def _push_block_addr(self, addr: int):
        for frame in self._block_addrs_stack:
            frame.add(addr)

    def _top_loop_block_addrs(self) -> set[int]:
        return self._block_addrs_stack[-1] if self._block_addrs_stack else set()

    def handle_CIfElse(self, obj):
        addr = obj.tags.get("ins_addr", None)
        if addr is not None:
            self._push_block_addr(addr)
        return super().handle_CIfElse(obj)

    def handle_CStatements(self, obj: CStatements):
        self._push_block_addr(obj.addr)
        return super().handle_CStatements(obj)

    def handle_CAssignment(self, obj: CAssignment):
        if isinstance(obj.lhs, CVariable) and obj.lhs.unified_variable is not None:
            self.var_values[obj.lhs.unified_variable.ident] = (
                obj.rhs.value if isinstance(obj.rhs, CConstant) and isinstance(obj.rhs.value, int) else None
            )
        return super().handle_CAssignment(obj)

    def handle_CWhileLoop(self, obj: CWhileLoop):
        self.result[obj.idx] = {
            "loop_type": "while",
        }

        self._enter_loop()
        ret = super().handle_CWhileLoop(obj)
        if obj.idx in self.result:
            self.result[obj.idx]["block_addrs"] = sorted(self._top_loop_block_addrs())
        self._leave_loop()
        return ret

    def handle_CDoWhileLoop(self, obj: CDoWhileLoop):
        cond, body = obj.condition, obj.body
        self.result[obj.idx] = {
            "loop_type": "do-while",
        }
        comp_ops = {"CmpNE", "CmpEQ", "CmpLT", "CmpLE", "CmpGT", "CmpGE"}
        if isinstance(cond, CBinaryOp) and cond.op in comp_ops:  # noqa:SIM102
            # parse the condition variable and constant if possible
            if isinstance(cond.lhs, CVariable):
                body_visitor = LoopBodyAssignmentCollector({cond.lhs.unified_variable.ident})
                body_visitor.handle(body)

                cond_node: BinOpNode | None = AssignmentToASTVisitor().handle(cond)
                if cond_node is not None and body_visitor.assignments:

                    assignments = body_visitor.assignments

                    # case 3:
                    # a = N
                    # do { b = a; c = b + A; a = c; } while (b ? M)
                    # we convert it into case 2 by eliminating c
                    if len(assignments) == 3:
                        assign1, assign2, assign3 = assignments  # pylint:disable=unbalanced-tuple-unpacking
                        if (
                            isinstance(assign2.rhs, BinOpNode)
                            and assign2.rhs.op in {"+", "-"}
                            and assign3.rhs == assign2.lhs
                            and isinstance(assign1, AssignmentNode)
                            and isinstance(assign1.rhs, VarNode)
                            and assign3.lhs == assign1.rhs
                            and assign1.lhs == cond_node.lhs
                        ):
                            assignments = [assign1, AssignmentNode(assign3.lhs, assign2.rhs)]

                    # case 2:
                    # a = N
                    # do { b = a; ...; a = b + A; } while (b ? M)
                    # we convert it into case 1
                    if len(assignments) == 2:
                        assign1, assign2 = assignments
                        if isinstance(assign2.rhs, BinOpNode) and assign2.rhs.op in {"+", "-"}:  # noqa:SIM102
                            if (
                                assign1.lhs == assign2.rhs.lhs
                                and assign1.rhs == assign2.lhs
                                and assign1.lhs == cond_node.lhs
                            ):
                                v = assign2.lhs

                                # update cond_node and assignments
                                new_const_value = cond_node.rhs.value
                                if assign2.rhs.op == "+":
                                    new_const_value += assign2.rhs.rhs.value
                                elif assign2.rhs.op == "-":
                                    new_const_value -= assign2.rhs.rhs.value
                                else:
                                    new_const_value = None
                                if new_const_value is not None:
                                    cond_node = BinOpNode(cond_node.op, v, ConstNode(new_const_value))
                                    # update assignments
                                    binop = BinOpNode(assign2.rhs.op, v, assign2.rhs.rhs)
                                    assignments = [AssignmentNode(v, binop)]

                    # case 1:
                    # a = N
                    # do { a = a + A; ,,, } while (a ? M)
                    init_value = None
                    if cond_node.lhs.var_ident in self.var_values:
                        init_value = self.var_values[cond_node.lhs.var_ident]
                    if init_value is not None and len(assignments) == 1:
                        assign = assignments[0]
                        step = None
                        if (
                            isinstance(assign.rhs, BinOpNode)
                            and assign.rhs.lhs.var_ident == cond_node.lhs.var_ident
                            and isinstance(assign.rhs.rhs, ConstNode)
                        ):
                            match assign.rhs.op:
                                case "+":
                                    step = assign.rhs.rhs.value
                                case "-":
                                    step = -assign.rhs.rhs.value
                                case _:
                                    step = None

                        if step is not None:
                            cond_op = cond_node.op
                            bound = cond_node.rhs.value

                            match cond_op:
                                case "==":
                                    max_iterations = 2 if bound == init_value + step else 1
                                case "!=":
                                    if (bound - init_value) % step == 0:
                                        max_iterations = (bound - init_value) // step
                                        if max_iterations <= 0:
                                            max_iterations = None
                                    else:
                                        max_iterations = None
                                case "<":
                                    max_iterations = (bound - init_value + step - 1) // step
                                    if max_iterations <= 0:
                                        max_iterations = None
                                case "<=":
                                    max_iterations = (bound - init_value + step - 2) // step
                                    if max_iterations <= 0:
                                        max_iterations = None
                                case ">":
                                    max_iterations = (bound - init_value) // step
                                    if max_iterations <= 0:
                                        max_iterations = None
                                case ">=":
                                    max_iterations = (bound - 1 - init_value) // step
                                    if max_iterations <= 0:
                                        max_iterations = None
                                case _:
                                    max_iterations = None

                            controls = LoopBodyControlStatementCollector()
                            controls.handle(obj)
                            fixed_iterations = not (controls.has_goto or controls.has_break or controls.has_continue)

                            self.result[obj.idx] = {
                                "loop_type": "do-while",
                                "loop_variable": cond_node.lhs.var_ident,
                                "initial_value": init_value,
                                "condition_op": cond_op,
                                "bound": bound,
                                "step": step,
                                "max_iterations": max_iterations,
                                "fixed_iterations": fixed_iterations,
                            }

        self._enter_loop()
        ret = super().handle_CDoWhileLoop(obj)
        if obj.idx in self.result:
            self.result[obj.idx]["block_addrs"] = sorted(self._top_loop_block_addrs())
        self._leave_loop()
        return ret

    def handle_CForLoop(self, obj: CForLoop):
        self.result[obj.idx] = {
            "loop_type": "for",
        }
        self._enter_loop()
        ret = super().handle_CForLoop(obj)
        if obj.idx in self.result:
            self.result[obj.idx]["block_addrs"] = sorted(self._top_loop_block_addrs())
        self._leave_loop()
        return ret


class LoopAnalysis(Analysis):
    """
    Analyze loop nodes in a structured C code representation and extract relevant information about the loop, including
    - Loop block addresses
    - Loop exits
    - Loop type
    - Loop condition
    - Max iterations
    - Fixed iterations
    """

    def __init__(self, cfunc: CFunction):
        super().__init__()

        self.cfunc = cfunc
        self.result: dict[str, dict] = {}

        self._analyze()

    def _analyze(self):
        visitor = LoopVisitor()
        visitor.handle(self.cfunc)
        self.result = visitor.result


register_analysis(LoopAnalysis, "LoopAnalysis")



# FIXME:
#     - Eliminate Vex references where possible
#     - Consider moving pieces of lifter classes to higher abstraction layer
#       to reduce duplication with Vex
#     - Fix default_exit_target
#     - Fix/remove NotImplementedError's
from __future__ import annotations

import logging
from typing import Any, TYPE_CHECKING
from collections.abc import Iterable, Sequence

import archinfo
import cle
import pypcode
from archinfo import ArchARM, ArchPcode
from cachetools import LRUCache

# FIXME: Reusing these errors from pyvex for compatibility. Eventually these
# should be refactored to use common error classes.
from pyvex.errors import PyVEXError, SkipStatementsError, LiftingException
from pyvex.expr import IRExpr, Const, U8, U16, U32, U64

from .behavior import BehaviorFactory
from angr.engines.engine import SimEngine
from angr.state_plugins.inspect import BP_AFTER, BP_BEFORE
from angr.sim_state import SimState
from angr.misc.ux import once
from angr.errors import SimEngineError, SimTranslationError, SimError
from angr import sim_options as o
from angr.block import DisassemblerBlock, DisassemblerInsn

if TYPE_CHECKING:
    from pypcode import PcodeOp, Context


l = logging.getLogger(__name__)

IRSB_MAX_SIZE = 400
IRSB_MAX_INST = 99
MAX_INSTRUCTIONS = 99999
MAX_BYTES = 5000


class ExitStatement:
    """
    This class exists to ease compatibility with CFGFast's processing of
    exit_statements. See _scan_irsb method.
    """

    __slots__ = ("dst", "jumpkind")

    dst: int | None
    jumpkind: str

    def __init__(self, dst: int | None, jumpkind: str):
        self.dst = dst
        self.jumpkind = jumpkind


class PcodeDisassemblerBlock(DisassemblerBlock):
    """
    Helper class to represent a block of disassembled target architecture
    instructions
    """


class PcodeDisassemblerInsn(DisassemblerInsn):
    """
    Helper class to represent a disassembled target architecture instruction
    """

    def __init__(self, pcode_insn):
        self.insn = pcode_insn

    @property
    def size(self) -> int:
        return self.insn.length

    @property
    def address(self) -> int:
        return self.insn.addr.offset

    @property
    def mnemonic(self) -> str:
        return self.insn.mnem

    @property
    def op_str(self) -> str:
        return self.insn.body


class IRSB:
    """
    IRSB stands for *Intermediate Representation Super-Block*. An IRSB in is a
    single-entry, multiple-exit code block.

    :ivar arch:             The architecture this block is lifted under
    :vartype arch:          :class:`archinfo.Arch`
    :ivar statements:       The statements in this block
    :vartype statements:    list of :class:`IRStmt`
    :ivar next:             The expression for the default exit target of this block
    :vartype next:          :class:`IRExpr`
    :ivar int offsIP:       The offset of the instruction pointer in the VEX guest state
    :ivar int stmts_used:   The number of statements in this IRSB
    :ivar str jumpkind:     The type of this block's default jump (call, boring, syscall, etc) as a VEX enum string
    :ivar bool direct_next: Whether this block ends with a direct (not indirect) jump or branch
    :ivar int size:         The size of this block in bytes
    :ivar int addr:         The address of this basic block, i.e. the address in the first IMark
    """

    __slots__ = (
        "_direct_next",
        "_disassembly",
        "_exit_statements",
        "_instruction_addresses",
        "_ops",
        "_size",
        "_statements",
        "addr",
        "arch",
        "behaviors",
        "const_vals",
        "data_refs",
        "default_exit_target",
        "jumpkind",
        "next",
    )

    _direct_next: bool | None
    _exit_statements: Sequence[tuple[int, int, ExitStatement]]
    _instruction_addresses: list[int] | None
    _ops: list[PcodeOp]  # FIXME: Merge into _statements
    _size: int | None
    _statements: Iterable  # Note: currently unused
    _disassembly: PcodeDisassemblerBlock | None
    addr: int
    arch: archinfo.Arch
    behaviors: BehaviorFactory | None
    data_refs: Sequence  # Note: currently unused
    const_vals: Sequence  # Note: currently unused
    default_exit_target: Any  # Note: currently used
    jumpkind: str | None
    next: IRExpr | None

    # The following constants shall match the defs in pyvex.h
    MAX_EXITS = 400
    MAX_DATA_REFS = 2000

    def __init__(
        self,
        data: str | bytes | None,
        mem_addr: int,
        arch: archinfo.Arch,
        max_inst: int | None = None,
        max_bytes: int | None = None,
        bytes_offset: int = 0,
        traceflags: int = 0,
        opt_level: int = 1,
        num_inst: int | None = None,
        num_bytes: int | None = None,
        strict_block_end: bool = False,
        skip_stmts: bool = False,
        collect_data_refs: bool = False,
    ) -> None:
        """
        :param data:                The bytes to lift. Can be either a string of bytes or a cffi buffer object.
                                    You may also pass None to initialize an empty IRSB.
        :param int mem_addr:        The address to lift the data at.
        :param arch:                The architecture to lift the data as.
        :param max_inst:            The maximum number of instructions to lift. (See note below)
        :param max_bytes:           The maximum number of bytes to use.
        :param num_inst:            Replaces max_inst if max_inst is None. If set to None as well, no instruction limit
                                    is used.
        :param num_bytes:           Replaces max_bytes if max_bytes is None. If set to None as well, no  byte limit is
                                    used.
        :param bytes_offset:        The offset into `data` to start lifting at. Note that for ARM THUMB mode, both
                                    `mem_addr` and `bytes_offset` must be odd (typically `bytes_offset` is set to 1).
        :param traceflags:          Unused by P-Code lifter
        :param opt_level:           Unused by P-Code lifter
        :param strict_block_end:    Unused by P-Code lifter

        .. note:: Explicitly specifying the number of instructions to lift (`max_inst`) may not always work
                  exactly as expected. For example, on MIPS, it is meaningless to lift a branch or jump
                  instruction without its delay slot. VEX attempts to Do The Right Thing by possibly decoding
                  fewer instructions than requested. Specifically, this means that lifting a branch or jump
                  on MIPS as a single instruction (`max_inst=1`) will result in an empty IRSB, and subsequent
                  attempts to run this block will raise `SimIRSBError('Empty IRSB passed to SimIRSB.')`.

        .. note:: If no instruction and byte limit is used, the lifter will continue lifting the block until the block
                  ends properly or until it runs out of data to lift.
        """
        if max_inst is None:
            max_inst = num_inst
        if max_bytes is None:
            max_bytes = num_bytes

        self._direct_next = None
        self._exit_statements = []
        self._instruction_addresses = None
        self._ops: list[PcodeOp] = []
        self._size = None
        self._statements = []
        self.addr = mem_addr
        self.arch = arch
        self.behaviors = None
        self.data_refs = ()
        self.const_vals = ()
        self.default_exit_target = None
        self.jumpkind = None
        self.next = None
        self._disassembly = None

        if data is not None:
            # This is the slower path (because we need to call _from_py() to copy the content in the returned IRSB to
            # the current IRSB instance. You should always call `lift()` directly. This method is kept for compatibility
            # concerns.
            irsb = lift(
                data,
                mem_addr,
                arch,
                max_bytes=max_bytes,
                max_inst=max_inst,
                bytes_offset=bytes_offset,
                opt_level=opt_level,
                traceflags=traceflags,
                strict_block_end=strict_block_end,
                skip_stmts=skip_stmts,
                collect_data_refs=collect_data_refs,
            )
            self._from_py(irsb)

    @staticmethod
    def empty_block(
        arch: archinfo.Arch,
        addr: int,
        statements: Sequence | None = None,
        nxt: int | None = None,
        tyenv=None,  # Unused, kept for compatibility
        jumpkind: str | None = None,
        direct_next: bool | None = None,
        size: int | None = None,
    ) -> IRSB:
        block = IRSB(None, addr, arch)
        block._set_attributes(statements, nxt, tyenv, jumpkind, direct_next, size=size)
        return block

    @property
    def has_statements(self) -> bool:
        return bool(self.statements is not None and self.statements)

    @property
    def exit_statements(self) -> Sequence[tuple[int, int, ExitStatement]]:
        return self._exit_statements

    def copy(self) -> IRSB:
        """
        Copy by creating an empty IRSB and then filling in the leftover
        attributes. Copy is made as deep as possible
        """
        new = IRSB.empty_block(
            arch=self.arch,
            addr=self.addr,
        )

        new._set_attributes(
            nxt=self.next,
            jumpkind=self.jumpkind,
            direct_next=self.direct_next,
            ops=self._ops[:],
        )

        return new

    def extend(self, extendwith: IRSB) -> IRSB:
        """
        Appends an irsb to the current irsb. The irsb that is appended is invalidated.
        The appended irsb's jumpkind and default exit are used.
        :param extendwith: The IRSB to append to this IRSB
        """
        self._set_attributes(
            nxt=extendwith.next,
            jumpkind=extendwith.jumpkind,
            direct_next=extendwith.direct_next,
            ops=self._ops + extendwith._ops,
        )

        self._disassembly = None
        return self

    def invalidate_direct_next(self) -> None:
        self._direct_next = None

    def pp(self) -> None:
        """
        Pretty-print the IRSB to stdout.
        """
        print(self._pp_str())

    def __repr__(self) -> str:
        return f"IRSB <0x{self.size:x} bytes, {self.instructions} ins., {self.arch}> at 0x{self.addr:x}"

    def __str__(self) -> str:
        return self._pp_str()

    #
    # simple properties useful for analysis
    #

    @property
    def tyenv(self):
        # FIXME: Here for compatibility. Remove.
        return None

    @property
    def stmts_used(self) -> int:
        if self.statements is None:
            return 0
        return len(self.statements)

    @property
    def offsIP(self) -> int | None:
        return self.arch.ip_offset

    @property
    def direct_next(self) -> bool:
        if self._direct_next is None:
            self._direct_next = self._is_defaultexit_direct_jump()
        return self._direct_next

    @property
    def expressions(self):
        """
        Return an iterator of all expressions contained in the IRSB.
        """
        raise NotImplementedError

    # FIXME: Rename this to num_instructions or something + fix pyvex IRSB.
    @property
    def instructions(self) -> int:
        """
        The number of instructions in this block
        """
        return len(self.instruction_addresses)

    @property
    def instruction_addresses(self) -> Sequence[int]:
        """
        Addresses of instructions in this block.
        """
        if self._instruction_addresses is None:
            self._instruction_addresses = []
            for op in self._ops:
                if op.opcode == pypcode.OpCode.IMARK:
                    for vn in op.inputs:
                        self._instruction_addresses.append(vn.offset)
        return self._instruction_addresses

    @property
    def size(self) -> int:
        """
        The size of this block, in bytes
        """
        assert self._size is not None
        return self._size

    @property
    def operations(self):
        """
        A list of all operations done by the IRSB, as libVEX enum names
        """
        raise NotImplementedError

    @property
    def all_constants(self):
        """
        Returns all constants in the block (including incrementing of the program counter) as
        :class:`pyvex.const.IRConst`.
        """
        raise NotImplementedError

    @property
    def constants(self):
        """
        The constants (excluding updates of the program counter) in the IRSB as :class:`pyvex.const.IRConst`.
        """
        raise NotImplementedError

    @property
    def constant_jump_targets(self):
        """
        A set of the static jump targets of the basic block.
        """
        exits = set()

        if self.exit_statements:
            for _, _, stmt in self.exit_statements:
                if stmt.dst is not None:
                    exits.add(stmt.dst)

        if self.next is not None:
            exits.add(self.next)

        return exits

    @property
    def constant_jump_targets_and_jumpkinds(self):
        """
        A dict of the static jump targets of the basic block to their jumpkind.
        """
        exits = {}

        if self.exit_statements:
            for _, _, stmt in self.exit_statements:
                if stmt.dst is not None:
                    exits[stmt.dst] = stmt.jumpkind

        if self.next is not None:
            exits[self.next] = self.jumpkind

        return exits

    @property
    def is_noop_block(self) -> bool:
        """
        Returns True if this block is a no-op block (i.e. it has no instructions and no jumps).
        """
        return not any(op.opcode != pypcode.OpCode.IMARK for op in self._ops)

    #
    # private methods
    #

    def _pp_str(self) -> str:
        """
        Return the pretty-printed IRSB.
        """
        sa = []
        sa.append("IRSB {")
        for i, op in enumerate(self._ops):
            if op.opcode == pypcode.OpCode.IMARK:
                for vn in op.inputs[:]:
                    sa.append(f"   {i:02d} | ------ {vn.offset:08x}, {vn.size} ------")
            else:
                sa.append(f"   {i:02d} | {pypcode.PcodePrettyPrinter.fmt_op(op)}")

        next_str = f"{self.next.con.value:x}" if isinstance(self.next, Const) else str(self.next)
        sa.append(f"   NEXT: {next_str}; {self.jumpkind}")
        sa.append("}")
        return "\n".join(sa)

    def _is_defaultexit_direct_jump(self) -> bool:
        """
        Checks if the default of this IRSB a direct jump or not.
        """
        if self.jumpkind not in ("Ijk_InvalICache", "Ijk_Boring", "Ijk_Call"):
            return False
        return self.default_exit_target is not None

    def _set_attributes(
        self: IRSB,
        statements: Iterable | None = None,
        nxt: int | Const | None = None,
        tyenv=None,  # Unused, kept for compatibility
        jumpkind: str | None = None,
        direct_next: bool | None = None,
        size: int | None = None,
        ops: list[PcodeOp] | None = None,
        instruction_addresses: list[int] | None = None,
        exit_statements: Sequence[tuple[int, int, ExitStatement]] | None = None,
        default_exit_target: Any = None,
    ) -> None:
        # pylint: disable=unused-argument
        self._statements = statements if statements is not None else []
        if isinstance(nxt, int):
            const_cls = {8: U8, 16: U16, 32: U32, 64: U64}[self.arch.bits]
            self.next = Const(const_cls(nxt))
        else:
            self.next = nxt
        self.jumpkind = jumpkind
        self._direct_next = direct_next
        self._size = size
        self._ops = ops or []
        self._instruction_addresses = instruction_addresses
        self._exit_statements = exit_statements or []
        self.default_exit_target = default_exit_target

    def _from_py(self, irsb: IRSB) -> None:
        self._set_attributes(
            irsb.statements,
            irsb.next,
            None,
            irsb.jumpkind,
            irsb.direct_next,
            irsb.size,
            ops=irsb._ops,
            instruction_addresses=irsb._instruction_addresses,
            exit_statements=irsb.exit_statements,
            default_exit_target=irsb.default_exit_target,
        )

    @property
    def statements(self) -> list:
        # FIXME: For compatibility, may want to implement Ist_IMark and
        # pyvex.IRStmt.Exit to ease analyses.
        l.debug("Returning empty statements list!")
        return []
        # return self._statements

    @property
    def disassembly(self) -> PcodeDisassemblerBlock:
        return self._disassembly


class Lifter:
    """
    A lifter is a class of methods for processing a block.

    :ivar data:             The bytes to lift as either a python string of bytes or a cffi buffer object.
    :ivar bytes_offset:     The offset into `data` to start lifting at.
    :ivar max_bytes:        The maximum number of bytes to lift. If set to None, no byte limit is used.
    :ivar max_inst:         The maximum number of instructions to lift. If set to None, no instruction limit is used.
    :ivar opt_level:        Unused by P-Code lifter
    :ivar traceflags:       Unused by P-Code lifter
    :ivar allow_arch_optimizations: Unused by P-Code lifter
    :ivar strict_block_end: Unused by P-Code lifter
    :ivar skip_stmts:       Unused by P-Code lifter
    """

    REQUIRE_DATA_C = False
    REQUIRE_DATA_PY = False

    __slots__ = (
        "addr",
        "allow_arch_optimizations",
        "arch",
        "bytes_offset",
        "collect_data_refs",
        "data",
        "irsb",
        "max_bytes",
        "max_inst",
        "opt_level",
        "skip_stmts",
        "strict_block_end",
        "traceflags",
    )

    data: str | bytes | None
    bytes_offset: int | None
    opt_level: int
    traceflags: int | None
    allow_arch_optimizations: bool | None
    strict_block_end: bool | None
    collect_data_refs: bool
    max_inst: int | None
    max_bytes: int | None
    skip_stmts: bool
    irsb: IRSB
    arch: archinfo.Arch
    addr: int

    def __init__(self, arch: archinfo.Arch, addr: int):
        self.arch = arch
        self.addr = addr
        self.data = None
        self.bytes_offset = None
        self.opt_level = 1
        self.traceflags = None
        self.allow_arch_optimizations = None
        self.strict_block_end = None
        self.collect_data_refs = False
        self.max_inst = None
        self.max_bytes = None
        self.skip_stmts = False
        self.irsb = None

    def _lift(
        self,
        data: str | bytes | None,
        bytes_offset: int | None = None,
        max_bytes: int | None = None,
        max_inst: int | None = None,
        opt_level: int = 1,
        traceflags: int | None = None,
        allow_arch_optimizations: bool | None = None,
        strict_block_end: bool | None = None,
        skip_stmts: bool = False,
        collect_data_refs: bool = False,
    ) -> IRSB:
        """
        Wrapper around the `lift` method on Lifters. Should not be overridden in child classes.

        :param data:                The bytes to lift as either a python string of bytes or a cffi buffer object.
        :param bytes_offset:        The offset into `data` to start lifting at.
        :param max_bytes:           The maximum number of bytes to lift. If set to None, no byte limit is used.
        :param max_inst:            The maximum number of instructions to lift. If set to None, no instruction limit is
                                    used.
        :param opt_level:           Unused by P-Code lifter
        :param traceflags:          Unused by P-Code lifter
        :param allow_arch_optimizations: Unused by P-Code lifter
        :param strict_block_end:    Unused by P-Code lifter
        :param skip_stmts:          Unused by P-Code lifter
        :param collect_data_refs:   Unused by P-Code lifter
        """
        irsb = IRSB.empty_block(self.arch, self.addr)
        self.data = data
        self.bytes_offset = bytes_offset
        self.opt_level = opt_level
        self.traceflags = traceflags
        self.allow_arch_optimizations = allow_arch_optimizations
        self.strict_block_end = strict_block_end
        self.collect_data_refs = collect_data_refs
        self.max_inst = max_inst
        self.max_bytes = max_bytes
        self.skip_stmts = skip_stmts
        self.irsb = irsb
        self.lift()
        return self.irsb

    def lift(self) -> None:
        """
        Lifts the data using the information passed into _lift. Should be overridden in child classes.

        Should set the lifted IRSB to self.irsb.
        If a lifter raises a LiftingException on the data, this signals that the lifter cannot lift this data and arch
        and the lifter is skipped.
        If a lifter can lift any amount of data, it should lift it and return the lifted block with a jumpkind of
        Ijk_NoDecode, signalling to pyvex that other lifters should be used on the undecodable data.

        """
        raise NotImplementedError


# pylint:disable=unused-argument
def lift(
    data: str | bytes | None,
    addr: int,
    arch: archinfo.Arch,
    max_bytes: int | None = None,
    max_inst: int | None = None,
    bytes_offset: int = 0,
    opt_level: int = 1,
    traceflags: int = 0,
    strict_block_end: bool = True,
    inner: bool = False,
    skip_stmts: bool = False,
    collect_data_refs: bool = False,
) -> IRSB:
    """
    Lift machine code in `data` to a P-code IRSB.

    If a lifter raises a LiftingException on the data, it is skipped.
    If it succeeds and returns a block with a jumpkind of Ijk_NoDecode, all of the lifters are tried on the rest
    of the data and if they work, their output is appended to the first block.

    :param arch:            The arch to lift the data as.
    :param addr:            The starting address of the block. Effects the IMarks.
    :param data:            The bytes to lift as either a python string of bytes or a cffi buffer object.
    :param max_bytes:       The maximum number of bytes to lift. If set to None, no byte limit is used.
    :param max_inst:        The maximum number of instructions to lift. If set to None, no instruction limit is used.
    :param bytes_offset:    The offset into `data` to start lifting at.
    :param opt_level:       Unused by P-Code lifter
    :param traceflags:      Unused by P-Code lifter

    .. note:: Explicitly specifying the number of instructions to lift (`max_inst`) may not always work
              exactly as expected. For example, on MIPS, it is meaningless to lift a branch or jump
              instruction without its delay slot. VEX attempts to Do The Right Thing by possibly decoding
              fewer instructions than requested. Specifically, this means that lifting a branch or jump
              on MIPS as a single instruction (`max_inst=1`) will result in an empty IRSB, and subsequent
              attempts to run this block will raise `SimIRSBError('Empty IRSB passed to SimIRSB.')`.

    .. note:: If no instruction and byte limit is used, the lifter will continue lifting the block until the block
              ends properly or until it runs out of data to lift.
    """
    if max_bytes is not None and max_bytes <= 0:
        raise PyVEXError("Cannot lift block with no data (max_bytes <= 0)")

    if not data:
        raise PyVEXError("Cannot lift block with no data (data is empty)")

    if isinstance(data, str):
        raise TypeError("Cannot pass unicode string as data to lifter")

    if isinstance(data, bytes):
        # py_data = data
        # c_data = None
        allow_arch_optimizations = False
    else:
        if max_bytes is None:
            raise PyVEXError("Cannot lift block with ffi pointer and no size (max_bytes is None)")
        # c_data = data
        # py_data = None
        allow_arch_optimizations = True

    # In order to attempt to preserve the property that
    # VEX lifts the same bytes to the same IR at all times when optimizations are disabled
    # we hack off all of VEX's non-IROpt optimizations when opt_level == -1.
    # This is intended to enable comparisons of the lifted IR between code that happens to be
    # found in different contexts.
    if opt_level < 0:
        allow_arch_optimizations = False
        opt_level = 0

    u_data = data
    try:
        final_irsb = PcodeLifter(arch, addr)._lift(
            u_data,
            bytes_offset,
            max_bytes,
            max_inst,
            opt_level,
            traceflags,
            allow_arch_optimizations,
            strict_block_end,
            skip_stmts,
            collect_data_refs,
        )
    except SkipStatementsError:
        assert skip_stmts is True
        final_irsb = PcodeLifter(arch, addr)._lift(
            u_data,
            bytes_offset,
            max_bytes,
            max_inst,
            opt_level,
            traceflags,
            allow_arch_optimizations,
            strict_block_end,
            skip_stmts=False,
            collect_data_refs=collect_data_refs,
        )
    except LiftingException as ex:
        l.debug("Lifting Exception: %s", ex)
        final_irsb = IRSB.empty_block(
            arch,
            addr,
            size=0,
            nxt=addr,
            jumpkind="Ijk_NoDecode",
        )
        final_irsb.invalidate_direct_next()
        return final_irsb

    if final_irsb.size > 0 and final_irsb.jumpkind == "Ijk_NoDecode":
        # We have decoded a few bytes before we hit an undecodeable instruction.

        # Determine if this is an intentional NoDecode, like the ud2 instruction on AMD64
        # FIXME:
        # nodecode_addr_expr = final_irsb.next
        # if type(nodecode_addr_expr) is Const:
        #     nodecode_addr = nodecode_addr_expr.con.value
        #     next_irsb_start_addr = addr + final_irsb.size
        #     if nodecode_addr != next_irsb_start_addr:
        #         # The last instruction of the IRSB has a non-zero length. This is an intentional NoDecode.
        #         # The very last instruction has been decoded
        #         final_irsb.jumpkind = "Ijk_NoDecode"
        #         final_irsb.next = final_irsb.next
        #         final_irsb.invalidate_direct_next()
        #         return final_irsb

        # Decode more bytes
        if skip_stmts:
            # When gymrat will be invoked, we will merge future basic blocks to the current basic block. In this case,
            # statements are usually required.
            # TODO: In the future, we may further optimize it to handle cases where getting statements in gymrat is not
            # TODO: required.
            return lift(
                data,
                addr,
                arch,
                max_bytes=max_bytes,
                max_inst=max_inst,
                bytes_offset=bytes_offset,
                opt_level=opt_level,
                traceflags=traceflags,
                strict_block_end=strict_block_end,
                skip_stmts=False,
                collect_data_refs=collect_data_refs,
            )

        next_addr = addr + final_irsb.size
        if max_bytes is not None:
            max_bytes -= final_irsb.size
        data_left = data[final_irsb.size :] if isinstance(data, (str, bytes, bytearray)) else data + final_irsb.size
        if max_inst is not None:
            max_inst -= final_irsb.instructions
        if (max_bytes is None or max_bytes > 0) and (max_inst is None or max_inst > 0) and data_left:
            more_irsb = lift(
                data_left,
                next_addr,
                arch,
                max_bytes=max_bytes,
                max_inst=max_inst,
                bytes_offset=bytes_offset,
                opt_level=opt_level,
                traceflags=traceflags,
                strict_block_end=strict_block_end,
                inner=True,
                skip_stmts=False,
                collect_data_refs=collect_data_refs,
            )
            if more_irsb.size:
                # Successfully decoded more bytes
                final_irsb.extend(more_irsb)
        elif max_bytes == 0:
            # We have no more bytes left. Mark the jumpkind of the IRSB as Ijk_Boring
            if final_irsb.size > 0 and final_irsb.jumpkind == "Ijk_NoDecode":
                final_irsb.jumpkind = "Ijk_Boring"
                const_cls = {8: U8, 16: U16, 32: U32, 64: U64}[arch.bits]
                final_irsb.next = Const(const_cls(final_irsb.addr + final_irsb.size))

    return final_irsb


class PcodeBasicBlockLifter:
    """
    Lifts basic blocks to P-code
    """

    context: Context
    behaviors: BehaviorFactory

    def __init__(self, arch: archinfo.Arch):
        if isinstance(arch, ArchPcode):
            langid = arch.name
        else:
            archinfo_to_lang_map = {
                "X86": "x86:LE:32:default",
                "AMD64": "x86:LE:64:default",
                "AVR8": "avr8:LE:16:atmega256",
            }
            if arch.name not in archinfo_to_lang_map:
                l.error("Unknown mapping of %s to pcode language id", arch.name)
                raise NotImplementedError
            langid = archinfo_to_lang_map[arch.name]

        self.context = pypcode.Context(langid)
        self.behaviors = BehaviorFactory()

    def lift(
        self,
        irsb: IRSB,
        baseaddr: int,
        data: bytes | bytearray,
        bytes_offset: int = 0,
        max_bytes: int | None = None,
        max_inst: int | None = None,
        branch_delay_slot: bool = False,
        is_sparc32: bool = False,
    ) -> None:
        assert irsb.addr == baseaddr
        assert bytes_offset < len(data)

        if max_bytes is None or max_bytes > MAX_BYTES:
            max_bytes = min(len(data) - bytes_offset, MAX_BYTES)
        if max_inst is None or max_inst > MAX_INSTRUCTIONS:
            max_inst = MAX_INSTRUCTIONS

        irsb.behaviors = self.behaviors  # FIXME

        # Translate
        sliced_data = bytes(data[bytes_offset : bytes_offset + max_bytes])

        if is_sparc32:
            # workaround to handle SPARC V8 decoding before having a SPARC V8 Sleigh file
            # replace all  jmpl xxx; rett xxx sequences with rett xxx; nop;
            nop_seq = b"\x01\x00\x00\x00"
            jmpl_seqs = [
                b"\x81\xc4\x40\x00",
                b"\x81\xc4\x80\x00",
            ]
            rett_seqs = [b"\x81\xcc\x80\x00", b"\x81\xcc\xa0\x04"]
            for jmpl_seq in jmpl_seqs:
                for rett_seq in rett_seqs:
                    seq = jmpl_seq + rett_seq
                    index = sliced_data.find(seq)
                    while index >= 0:
                        sliced_data = sliced_data[:index] + rett_seq + nop_seq + sliced_data[index + 8 :]
                        index = sliced_data.find(seq)

        sliced_data = bytes(sliced_data)

        # Post-process block to mark exits and next block
        next_block = None
        irsb._instruction_addresses = []
        fallthru_addr = irsb.addr

        try:
            translation = self.context.translate(
                sliced_data,
                irsb.addr,
                max_instructions=max_inst,
                max_bytes=max_bytes,
                flags=pypcode.TranslateFlags.BB_TERMINATING,
            )
            irsb._ops = translation.ops

            last_decode_addr = irsb.addr
            last_imark_idx = 0
            for op_idx, op in enumerate(irsb._ops):
                if op.opcode == pypcode.OpCode.IMARK:
                    irsb._instruction_addresses.extend([vn.offset for vn in op.inputs])
                    last_decode_addr = op.inputs[0].offset
                    fallthru_addr = op.inputs[-1].offset + op.inputs[-1].size
                    last_imark_idx = op_idx
                    continue

                if op.opcode in {pypcode.OpCode.BRANCH, pypcode.OpCode.CBRANCH} and op.inputs[0].space.name == "const":
                    # P-code relative branch (op_idx + op.inputs[0].offset)
                    # Note: We only model these in execution
                    continue

                if op.opcode == pypcode.OpCode.CBRANCH:
                    irsb._exit_statements.append(
                        (last_decode_addr, op_idx - last_imark_idx, ExitStatement(op.inputs[0].offset, "Ijk_Boring"))
                    )
                elif op.opcode == pypcode.OpCode.BRANCH:
                    if next_block is None:
                        next_block = (op.inputs[0].offset, "Ijk_Boring")
                elif op.opcode == pypcode.OpCode.BRANCHIND:
                    if next_block is None:
                        next_block = (None, "Ijk_Boring")
                elif op.opcode == pypcode.OpCode.CALL:
                    if next_block is None:
                        next_block = (op.inputs[0].offset, "Ijk_Call")
                elif op.opcode == pypcode.OpCode.CALLIND:
                    if next_block is None:
                        next_block = (None, "Ijk_Call")
                elif op.opcode == pypcode.OpCode.RETURN and next_block is None:
                    next_block = (None, "Ijk_Ret")

            # FIXME: Do this lazily
            disasm = self.context.disassemble(
                sliced_data,
                irsb.addr,
                max_instructions=max_inst,
                max_bytes=fallthru_addr - irsb.addr,
            )
            irsb._disassembly = PcodeDisassemblerBlock(
                addr=irsb.addr,
                insns=[PcodeDisassemblerInsn(ins) for ins in disasm.instructions],
                thumb=False,
                arch=irsb.arch,
            )

        except (pypcode.BadDataError, pypcode.UnimplError):
            next_block = (fallthru_addr, "Ijk_NoDecode")
        except (pypcode.LowlevelError, IndexError):
            # FIXME:
            # - IndexError: Give more data
            # - pypcode.LowlevelError: Sometimes a decoding failure
            next_block = (irsb.addr, "Ijk_NoDecode")

        if next_block is None:
            next_block = (fallthru_addr, "Ijk_Boring")

        irsb._size = fallthru_addr - irsb.addr
        const_cls = {8: U8, 16: U16, 32: U32, 64: U64}[irsb.arch.bits]
        irsb.next = Const(const_cls(next_block[0])) if next_block[0] is not None else None
        irsb.jumpkind = next_block[1]


class PcodeLifter(Lifter):
    """
    Handles calling into pypcode to lift a block
    """

    _lifter_cache = {}

    def lift(self) -> None:
        if self.arch not in PcodeLifter._lifter_cache:
            PcodeLifter._lifter_cache[self.arch] = PcodeBasicBlockLifter(self.arch)
        lifter = PcodeLifter._lifter_cache[self.arch]
        lifter.lift(
            self.irsb,
            self.addr,
            self.data,
            bytes_offset=self.bytes_offset,
            max_inst=self.max_inst,
            max_bytes=self.max_bytes,
            branch_delay_slot=self.arch.branch_delay_slot,
            is_sparc32="sparc:" in self.arch.name and self.arch.bits == 32,
        )

        if self.irsb.size == 0:
            l.debug("raising lifting exception")
            raise LiftingException(f"pypcode: could not decode any instructions @ 0x{self.addr:x}")


class PcodeLifterEngineMixin(SimEngine):
    """
    Lifter mixin to lift from machine code to P-Code.
    """

    def __init__(
        self,
        project=None,
        use_cache: bool | None = None,
        cache_size: int = 50000,
        default_opt_level: int = 1,
        selfmodifying_code: bool | None = None,
        single_step: bool = False,
        default_strict_block_end: bool = False,
        **kwargs,
    ):
        super().__init__(project, **kwargs)

        self._use_cache = use_cache
        self._default_opt_level = default_opt_level
        self._cache_size = cache_size
        self.selfmodifying_code = selfmodifying_code
        self._single_step = single_step
        self.default_strict_block_end = default_strict_block_end

        if self._use_cache is None:
            if self.project is not None:
                self._use_cache = self.project._translation_cache
            else:
                self._use_cache = False
        if self.selfmodifying_code is None:
            if self.project is not None:
                self.selfmodifying_code = self.project.selfmodifying_code
            else:
                self.selfmodifying_code = False

        # block cache
        self._block_cache = None
        self._block_cache_hits = 0
        self._block_cache_misses = 0
        self._initialize_block_cache()

    def _initialize_block_cache(self) -> None:
        self._block_cache = LRUCache(maxsize=self._cache_size)
        self._block_cache_hits = 0
        self._block_cache_misses = 0

    def clear_cache(self) -> None:
        self._block_cache = LRUCache(maxsize=self._cache_size)
        self._block_cache_hits = 0
        self._block_cache_misses = 0

    # FIXME: Consider moving to higher abstraction layer to reduce duplication with vex
    def lift_vex(
        self,
        addr: int | None = None,
        state: SimState | None = None,
        clemory: cle.Clemory | cle.ClemoryReadOnlyView | None = None,
        insn_bytes: bytes | None = None,
        arch: archinfo.Arch | None = None,
        size: int | None = None,
        num_inst: int | None = None,
        traceflags: int = 0,
        thumb: bool = False,
        extra_stop_points: Iterable[int] | None = None,
        opt_level: int | None = None,
        strict_block_end: bool | None = None,
        skip_stmts: bool = False,
        collect_data_refs: bool = False,
        load_from_ro_regions: bool = False,
        cross_insn_opt: bool | None = None,
        const_prop: bool | None = None,
    ) -> IRSB:
        """
        Temporary compatibility interface for integration with block code.
        """
        return self.lift_pcode(
            addr,
            state,
            clemory,
            insn_bytes,
            arch,
            size,
            num_inst,
            traceflags,
            thumb,
            extra_stop_points,
            opt_level,
            strict_block_end,
            skip_stmts,
            collect_data_refs,
            load_from_ro_regions,
            cross_insn_opt,
            const_prop,
        )

    def lift_pcode(
        self,
        addr: int | None = None,
        state: SimState | None = None,
        clemory: cle.Clemory | cle.ClemoryReadOnlyView | None = None,
        insn_bytes: bytes | None = None,
        arch: archinfo.Arch | None = None,
        size: int | None = None,
        num_inst: int | None = None,
        traceflags: int = 0,
        thumb: bool = False,
        extra_stop_points: Iterable[int] | None = None,
        opt_level: int | None = None,
        strict_block_end: bool | None = None,
        skip_stmts: bool = False,
        collect_data_refs: bool = False,
        load_from_ro_regions: bool = False,
        cross_insn_opt: bool | None = None,
        const_prop: bool | None = None,
    ) -> IRSB:
        """
        Lift an IRSB.

        There are many possible valid sets of parameters. You at the very least must pass some
        source of data, some source of an architecture, and some source of an address.

        Sources of data in order of priority: insn_bytes, clemory, state

        Sources of an address, in order of priority: addr, state

        Sources of an architecture, in order of priority: arch, clemory, state

        :param state:           A state to use as a data source.
        :param clemory:         A cle.memory.Clemory object to use as a data source.
        :param addr:            The address at which to start the block.
        :param thumb:           Whether the block should be lifted in ARM's THUMB mode.
        :param opt_level:       Unused for P-Code lifter
        :param insn_bytes:      A string of bytes to use as a data source.
        :param size:            The maximum size of the block, in bytes.
        :param num_inst:        The maximum number of instructions.
        :param traceflags:      Unused by P-Code lifter
        :param strict_block_end: Unused by P-Code lifter
        :param load_from_ro_regions: Unused by P-Code lifter
        """
        if cross_insn_opt:
            l.debug("cross_insn_opt is ignored for p-code lifter")
        if const_prop:
            l.debug("const_prop is ignored for p-code lifter")
        if load_from_ro_regions:
            l.debug("load_from_ro_regions is ignored for p-code lifter")

        # phase 0: sanity check
        if not state and not clemory and not insn_bytes:
            raise ValueError("Must provide state or clemory or insn_bytes!")
        if not state and not clemory and not arch:
            raise ValueError("Must provide state or clemory or arch!")
        if addr is None and not state:
            raise ValueError("Must provide state or addr!")
        if arch is None:
            arch = clemory._arch if clemory else state.arch
        if arch.name.startswith("MIPS") and self._single_step:
            l.error("Cannot specify single-stepping on MIPS.")
            self._single_step = False

        # phase 1: parameter defaults
        if addr is None:
            assert state is not None
            addr = state.solver.eval(state._ip)
        if size is not None:
            size = min(size, IRSB_MAX_SIZE)
        if size is None:
            size = IRSB_MAX_SIZE
        if num_inst is not None:
            num_inst = min(num_inst, IRSB_MAX_INST)
        if num_inst is None and self._single_step:
            num_inst = 1
        if opt_level is None:
            opt_level = 1 if state and o.OPTIMIZE_IR in state.options else self._default_opt_level
        if strict_block_end is None:
            strict_block_end = self.default_strict_block_end
        if self.selfmodifying_code and opt_level > 0:
            if once("vex-engine-smc-opt-warning"):
                l.warning(
                    "Self-modifying code is not always correctly optimized by"
                    " PyVEX. To guarantee correctness, VEX optimizations have been"
                    " disabled."
                )
            opt_level = 0
            assert state is not None
            if state and o.OPTIMIZE_IR in state.options:
                state.options.remove(o.OPTIMIZE_IR)
        if skip_stmts is not True:
            skip_stmts = False

        have_patches = self.project and self.project.kb.patches.items()
        use_cache = self._use_cache and not (skip_stmts or collect_data_refs or have_patches or state)

        # phase 2: thumb normalization
        thumb = int(thumb)
        if isinstance(arch, ArchARM):
            if addr % 2 == 1:
                thumb = 1
            if thumb:
                addr &= ~1
        elif thumb:
            l.error("thumb=True passed on non-arm architecture!")
            thumb = 0

        # phase 3: check cache
        cache_key = None
        if use_cache:
            cache_key = (
                addr,
                insn_bytes,
                size,
                num_inst,
                thumb,
                opt_level,
                strict_block_end,
            )
            if cache_key in self._block_cache:
                self._block_cache_hits += 1
                irsb = self._block_cache[cache_key]
                stop_point = self._first_stoppoint(irsb, extra_stop_points)
                if stop_point is None:
                    return irsb
                size = stop_point - addr
                # check the cache again
                cache_key = (
                    addr,
                    insn_bytes,
                    size,
                    num_inst,
                    thumb,
                    opt_level,
                    strict_block_end,
                )
                if cache_key in self._block_cache:
                    self._block_cache_hits += 1
                    return self._block_cache[cache_key]
                self._block_cache_misses += 1
            else:
                # a special case: `size` is used as the maximum allowed size
                tmp_cache_key = (
                    addr,
                    insn_bytes,
                    IRSB_MAX_SIZE,
                    num_inst,
                    thumb,
                    opt_level,
                    strict_block_end,
                )
                try:
                    irsb = self._block_cache[tmp_cache_key]
                    if irsb.size <= size:
                        self._block_cache_hits += 1
                        return self._block_cache[tmp_cache_key]
                except KeyError:
                    self._block_cache_misses += 1

        # vex_lift breakpoints only triggered when the cache isn't used
        if state:
            state._inspect("vex_lift", BP_BEFORE, mem_read_address=addr, mem_read_length=size)

        # phase 4: get bytes
        if insn_bytes is not None:
            buff, size = insn_bytes, len(insn_bytes)
        else:
            buff, size, _ = self._load_bytes(addr, size, state, clemory)

        if not buff or size == 0:
            raise SimEngineError(f"No bytes in memory for block starting at {addr:#x}.")

        # phase 5: lift to pcode
        l.debug("Creating pcode.IRSB of arch %s at %#x", arch.name, addr)
        try:
            for subphase in range(2):
                irsb = lift(
                    buff,
                    addr + thumb,
                    arch,
                    max_bytes=size,
                    max_inst=num_inst,
                    bytes_offset=thumb,
                    traceflags=traceflags,
                    opt_level=opt_level,
                    strict_block_end=strict_block_end,
                    skip_stmts=skip_stmts,
                    collect_data_refs=collect_data_refs,
                )

                if subphase == 0 and irsb.statements is not None:
                    # check for possible stop points
                    stop_point = self._first_stoppoint(irsb, extra_stop_points)
                    if stop_point is not None:
                        size = stop_point - addr
                        continue

                if use_cache:
                    self._block_cache[cache_key] = irsb
                if state:
                    state._inspect(
                        "vex_lift",
                        BP_AFTER,
                        mem_read_address=addr,
                        mem_read_length=size,
                    )
                return irsb

            raise SimEngineError("Unreachable code reached")
        # phase x: error handling
        except PyVEXError as e:
            l.debug("Translation error at %#x", addr)
            raise SimTranslationError("Unable to translate bytecode") from e

    def _load_bytes(
        self,
        addr: int,
        max_size: int,
        state: SimState | None = None,
        clemory: cle.Clemory | cle.ClemoryReadOnlyView | None = None,
    ) -> tuple[bytes, int, int]:
        if clemory is None and state is None:
            raise SimEngineError("state and clemory cannot both be None in _load_bytes().")

        buff, size, offset = b"", 0, 0

        # XXX: Prioritize loading from patched state, if we have patches
        have_patches = self.project and self.project.kb.patches.items()
        if state is None and have_patches:
            state = self.project.kb.patches.patched_entry_state

        load_from_state = self.selfmodifying_code or have_patches

        # skip loading from the clemory if we're using the ultra page
        # TODO: is this a good change? it neuters lookback optimizations
        # we can try concrete loading the full page but that has drawbacks too...
        # if state is not None and issubclass(getattr(state.memory, 'PAGE_TYPE', object), UltraPage):
        #    load_from_state = True

        # Load from the clemory if we can
        if not load_from_state or not state:
            if isinstance(clemory, (cle.Clemory, cle.ClemoryReadOnlyView)):
                try:
                    start, backer = next(clemory.backers(addr))
                except StopIteration:
                    pass
                else:
                    if start <= addr:
                        offset = addr - start
                        if isinstance(backer, (bytes, bytearray)):
                            avail = len(backer) - offset
                            size = min(avail, max_size)
                            buff = backer[offset : offset + size]
                        elif isinstance(backer, list):
                            raise SimTranslationError(
                                "Cannot lift block for arch with strange byte width. If you think you ought to be able "
                                "to, open an issue."
                            )
                        else:
                            raise TypeError(f"Unsupported backer type {type(backer)}.")
            elif state:
                if state.memory.SUPPORTS_CONCRETE_LOAD:
                    buff = state.memory.concrete_load(addr, max_size)
                else:
                    buff = state.solver.eval(state.memory.load(addr, max_size, inspect=False), cast_to=bytes)
                size = len(buff)

        # If that didn't work and if load_from_state is enabled, try to load from the state
        if load_from_state and state and size == 0:
            if state.memory.SUPPORTS_CONCRETE_LOAD:
                buff = state.memory.concrete_load(addr, max_size)
            else:
                buff = state.solver.eval(state.memory.load(addr, max_size, inspect=False), cast_to=bytes)
            size = len(buff)
            if self.selfmodifying_code and size < min(max_size, 10):  # arbitrary metric for doing the slow path
                l.debug("SMC slow path")
                buff_lst = []
                symbolic_warned = False
                for i in range(max_size):
                    try:
                        byte = state.memory.load(addr + i, 1, inspect=False)
                        if byte.symbolic and not symbolic_warned:
                            symbolic_warned = True
                            l.warning("Executing symbolic code at %#x", addr + i)
                        buff_lst.append(state.solver.eval(byte))
                    except SimError:
                        break

                buff = bytes(buff_lst)
                size = len(buff)

        size = min(max_size, size)
        return buff, size, offset

    def _first_stoppoint(self, irsb: IRSB, extra_stop_points: Sequence[int] | None = None) -> int | None:
        """
        Enumerate the imarks in the block. If any of them (after the first one) are at a stop point, returns the address
        of the stop point. None is returned otherwise.
        """
        if extra_stop_points is None and self.project is None:
            return None

        first_imark = True
        for addr in irsb.instruction_addresses:
            if not first_imark and self.__is_stop_point(addr, extra_stop_points):
                # could this part be moved by pyvex?
                return addr
            first_imark = False
        return None

    def __is_stop_point(self, addr: int, extra_stop_points: Sequence[int] | None = None) -> bool:
        return bool(
            (self.project is not None and addr in self.project._sim_procedures)
            or (extra_stop_points is not None and addr in extra_stop_points)
        )

    def __getstate__(self):
        ostate = super().__getstate__()
        s = {
            "_use_cache": self._use_cache,
            "_default_opt_level": self._default_opt_level,
            "selfmodifying_code": self.selfmodifying_code,
            "_single_step": self._single_step,
            "_cache_size": self._cache_size,
            "default_strict_block_end": self.default_strict_block_end,
        }

        return (s, ostate)

    def __setstate__(self, state):
        s, ostate = state
        self._use_cache = s["_use_cache"]
        self._default_opt_level = s["_default_opt_level"]
        self.selfmodifying_code = s["selfmodifying_code"]
        self._single_step = s["_single_step"]
        self._cache_size = s["_cache_size"]
        self.default_strict_block_end = s["default_strict_block_end"]

        # rebuild block cache
        self._initialize_block_cache()
        super().__setstate__(ostate)



from __future__ import annotations
import angr


class getchar(angr.SimProcedure):
    # pylint: disable=arguments-differ
    def run(self):
        fgetc = angr.SIM_PROCEDURES["libc"]["fgetc"]
        stdin = self.state.posix.get_fd(0)
        return self.inline_call(fgetc, 0, simfd=stdin).ret_expr


getchar_unlocked = getchar



from __future__ import annotations
import logging

import archinfo
from archinfo.arch_arm import is_arm_arch, ArchARMHF, ArchARMCortexM

from angr.calling_conventions import SimCC

l = logging.getLogger(__name__)


def is_sane_register_variable(
    arch: archinfo.Arch, reg_offset: int, reg_size: int, def_cc: SimCC | type[SimCC] | None = None
) -> bool:
    """
    Filters all registers that are surly not members of function arguments.
    This can be seen as a workaround, since VariableRecoveryFast sometimes gives input variables of cc_ndep (which
    is a VEX-specific register) :-(

    :param reg_offset:  The register offset.
    :param reg_size:    The register size.
    :return:            True if it is an acceptable function argument, False otherwise.
    :rtype:             bool
    """

    arch_name = arch.name
    if ":" in arch_name:
        # for pcode architectures, we only leave registers that are known to be used as input arguments
        if def_cc is not None:
            return arch.translate_register_name(reg_offset, size=reg_size) in def_cc.ARG_REGS
        return True

    # VEX
    if arch_name == "AARCH64":
        return 16 <= reg_offset < 80  # x0-x7

    if arch_name == "AMD64":
        # TODO is rbx ever a register?
        return 24 <= reg_offset < 40 or 64 <= reg_offset < 104  # rcx, rdx  # rsi, rdi, r8, r9, r10
        # 224 <= reg_offset < 480)  # xmm0-xmm7

    if is_arm_arch(arch):
        if isinstance(arch, (ArchARMHF, ArchARMCortexM)):
            return 8 <= reg_offset < 24 or 128 <= reg_offset < 160  # r0 - 32  # s0 - s7, or d0 - d4
        return 8 <= reg_offset < 24  # r0-r3

    if arch_name == "MIPS32":
        return 24 <= reg_offset < 40  # a0-a3

    if arch_name == "MIPS64":
        return 48 <= reg_offset < 80 or 112 <= reg_offset < 208  # a0-a3 or t4-t7

    if arch_name == "PPC32":
        return 28 <= reg_offset < 60  # r3-r10

    if arch_name == "X86":
        return 8 <= reg_offset < 24 or 160 <= reg_offset < 288  # eax, ebx, ecx, edx  # xmm0-xmm7

    l.critical("Unsupported architecture %s.", arch.name)
    return True



from __future__ import annotations

from .boolean_counter import BooleanCounter
from .call_counter import AILBlockCallCounter
from .seq_cf_structure_counter import ControlFlowStructureCounter
from .expression_counters import SingleExpressionCounter, RegisterExpressionCounter, OperatorCounter


__all__ = (
    "AILBlockCallCounter",
    "BooleanCounter",
    "ControlFlowStructureCounter",
    "OperatorCounter",
    "RegisterExpressionCounter",
    "SingleExpressionCounter",
)



from __future__ import annotations
import angr

# pylint: disable=arguments-differ,unused-argument,no-self-use,inconsistent-return-statements


class pthread_create(angr.SimProcedure):
    """
    Simulates the new thread as an equally viable branch of symbolic execution.
    """

    ADDS_EXITS = True

    # pylint: disable=unused-argument,arguments-differ
    def run(self, thread, attr, start_routine, arg):
        self.call(start_routine, (arg,), "terminate_thread", prototype="void *start_routine(void*)")
        return 0

    def terminate_thread(self, thread, attr, start_routine, arg):
        self.exit(0)

    def static_exits(self, blocks, **kwargs):
        # Execute those blocks with a blank state, and then dump the arguments
        blank_state = angr.SimState(project=self.project, mode="fastpath", cle_memory_backer=self.project.loader.memory)

        # Execute each block
        state = blank_state
        for b in blocks:
            irsb = self.project.factory.default_engine.process(state, b, force_addr=b.addr)
            if irsb.successors:
                state = irsb.successors[0]
            else:
                break

        callfunc = self.cc.get_args(state, self.prototype)[2]
        retaddr = state.memory.load(state.regs.sp, size=self.arch.bytes)

        return [
            {"address": callfunc, "jumpkind": "Ijk_Call", "namehint": "thread_entry"},
            {"address": retaddr, "jumpkind": "Ijk_Ret", "namehint": None},
        ]


class pthread_cond_signal(angr.SimProcedure):
    """
    A no-op.
    """

    def run(self, arg):
        pass


class pthread_mutex_lock(angr.SimProcedure):
    """
    Always returns 0 (SUCCESS).
    """

    def run(self, arg):
        return 0


class pthread_mutex_unlock(angr.SimProcedure):
    """
    Always returns 0 (SUCCESS).
    """

    def run(self, arg):
        return 0


class pthread_once(angr.SimProcedure):
    def run(self, control, func):
        controlword = self.state.mem[control].char.resolved
        if (controlword & 2).symbolic:
            raise angr.errors.SimProcedureError("Cannot handle symbolic control data for pthread_once")
        if self.state.solver.is_true(controlword & 2 != 0):
            return 0

        controlword |= 2
        self.state.mem[control].char = controlword
        self.call(func, (), "retsite", prototype="void x()")
        return None

    def retsite(self, control, func):
        return 0

    # TODO: static exits



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase

default_plugins = {}


class KnowledgeBasePlugin:
    def __init__(self, kb: KnowledgeBase):
        # This call is needed, because some Plugins like TypesStore inherit from KnowledgeBasePlugin and something else
        # So we still need to call the constructor of the other class.
        super().__init__()
        self._kb = kb

    def copy(self):
        raise NotImplementedError

    @staticmethod
    def register_default(name, cls):
        if name in default_plugins:
            raise Exception(f"{default_plugins[name]} is already set as the default for {name}")
        default_plugins[name] = cls



from __future__ import annotations
import logging

from angr.sim_type import SimTypeFunction, SimTypeInt, SimTypePointer, SimTypeChar, SimTypeBottom

from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


gnulib = SimLibrary()
gnulib.set_library_names("gnulib.so")
gnulib.add_all_from_dict(P["gnulib"])
gnulib.set_non_returning("xstrtol_fatal", "xalloc_die")


#
# parsed function prototypes
#

_gnulib_decls = {
    # pylint: disable=line-too-long
    # void xstrtol_fatal(enum strtol_error err, int opt_idx, char c, struct option const *long_options, char const *arg)
    # pylint: disable=line-too-long
    "xstrtol_fatal": SimTypeFunction(
        [
            SimTypeInt(signed=True),
            SimTypeInt(signed=True),
            SimTypeChar(),
            SimTypePointer(SimTypeInt(), offset=0),
            SimTypePointer(SimTypeChar(), offset=0),
        ],
        SimTypeBottom(label="void"),
        arg_names=["err", "opt_idx", "c", "long_options", "arg"],
    ),
    # void xalloc_die()
    # pylint: disable=line-too-long
    "xalloc_die": SimTypeFunction([], SimTypeBottom(label="void")),
}



from __future__ import annotations
import contextlib
from typing import TYPE_CHECKING
import logging

from angr import sim_options as o
from angr import BP, BP_AFTER
from angr.errors import (
    AngrUnsupportedSyscallError,
    SimOperationError,
    SimError,
)

from .resolver import IndirectJumpResolver
from .constant_value_manager import ConstantValueManager

if TYPE_CHECKING:
    from angr import Block
    from angr.engines import SimSuccessors
    from angr.sim_state import SimState
    from angr.sim_procedure import SimProcedure


_l = logging.getLogger(name=__name__)


class SyscallResolver(IndirectJumpResolver):
    """
    Resolve syscalls to SimProcedures.
    """

    def __init__(self, project):
        super().__init__(project, timeless=True)

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        return jumpkind.startswith("Ijk_Sys")

    def resolve(  # pylint:disable=unused-argument
        self, cfg, addr: int, func_addr: int, block: Block, jumpkind: str, func_graph_complete: bool = True, **kwargs
    ):
        stub = self._resolve_syscall_to_stub(cfg, addr, func_addr, block)
        return (True, [stub.addr]) if stub else (False, [])

    def _resolve_syscall_to_stub(self, cfg, addr: int, func_addr: int, block: Block) -> SimProcedure | None:
        if not cfg.functions.contains_addr(func_addr):
            return None
        func = cfg.functions.get_by_addr(func_addr)

        cv_manager = ConstantValueManager(self.project, cfg.kb, func, addr)
        constant_value_reg_read_bp = BP(when=BP_AFTER, enabled=True, action=cv_manager.reg_read_callback)

        state = self.project.factory.blank_state(
            mode="fastpath",
            addr=block.addr,
            add_options={o.SYMBOL_FILL_UNCONSTRAINED_MEMORY, o.SYMBOL_FILL_UNCONSTRAINED_REGISTERS},
        )
        state.inspect.add_breakpoint("reg_read", constant_value_reg_read_bp)

        successors = self._simulate_block_with_resilience(state)
        if successors:
            state = self._get_syscall_state_from_successors(successors)
            if state:
                with contextlib.suppress(AngrUnsupportedSyscallError):
                    return self.project.simos.syscall(state)
        return None

    def _simulate_block_with_resilience(self, state: SimState) -> SimSuccessors | None:
        """
        Execute a basic block with "On Error Resume Next". Give up when there is no way moving forward.
        """

        stmt_idx = 0
        successors = None  # make PyCharm's linting happy

        while True:
            try:
                successors = self.project.factory.successors(state, skip_stmts=stmt_idx)
                break
            except SimOperationError:
                stmt_idx += 1
                continue
            except SimError:
                return None

        return successors

    @staticmethod
    def _get_syscall_state_from_successors(successors: SimSuccessors) -> SimState | None:
        for state in successors.flat_successors:
            if state.history.jumpkind and state.history.jumpkind.startswith("Ijk_Sys"):
                return state
        return None
